# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

import time
import pytest

import unidist
from unidist.config import Backend
from unidist.core.base.common import BackendName
from .utils import (
    assert_equal,
    catch_exception,
    task,
    task_return_none,
    task_multiple_returns_default,
    task_multiple_returns,
)

unidist.init()


def test_remote_get():
    assert_equal(task.remote(5), 25)


def test_chaining_value():
    object_ref = unidist.put(6)
    object_refs = [task.remote(object_ref) for _ in range(4)]
    assert_equal(object_refs, [36, 36, 36, 36])


def test_num_returns_decorator_options():
    object_ref0, object_ref1 = task_multiple_returns_default.remote(5)
    assert_equal(object_ref0, 5)
    assert_equal(object_ref1, 25)


def test_num_returns_options():
    object_ref0, object_ref1 = task_multiple_returns.options(num_returns=2).remote(5)
    assert_equal(object_ref0, 5)
    assert_equal(object_ref1, 25)


def test_put_lambda():
    object_ref0 = unidist.put(lambda x: x * x)

    @unidist.remote
    def foo(f):
        return f(3)

    object_ref1 = foo.remote(object_ref0)
    assert_equal(object_ref1, 9)


def test_delayed_object_ref():
    @unidist.remote
    def foo():
        time.sleep(3)
        return 7

    object_ref0 = foo.remote()
    object_ref1 = task.remote(object_ref0)
    assert_equal(object_ref1, 49)


def test_num_returns_zero():
    @unidist.remote
    def foo():
        pass

    assert_equal(foo.options(num_returns=0).remote(), None)


def test_exception():
    @unidist.remote
    def foo(x):
        return x / x

    catch_exception(foo.remote(0), ZeroDivisionError)


def test_return_none():
    assert_equal(task_return_none.remote(), None)


@pytest.mark.skipif(
    Backend.get() == BackendName.PYMP,
    reason="Run of a remote task inside of another one is not implemented yet for pymp",
)
def test_internal_remote():
    @unidist.remote
    def foo(x):
        o_r = task.remote(x)
        return unidist.get(o_r) + 10

    assert_equal(foo.remote(5), 35)


@pytest.mark.skipif(
    Backend.get() == BackendName.PYMP,
    reason="Serialization of `dict_keys` is not properly implemented yet for pymp",
)
def test_serialize_dict_keys():
    @unidist.remote
    def f(keys):
        return keys

    dict_obj = {"1": 1, "2": 2, "3": 3, "4": 4, "5": 5}

    assert_equal(f.remote(dict_obj.keys()), dict_obj.keys())


def test_serialize_dict_with_tuple_key():
    @unidist.remote
    def f(params):
        return params

    data = {
        "a": [0, 1, 2, 3],
        "b": {("a", "b", "c", "d"): "value"},
    }

    assert_equal(f.remote(data), data)


@pytest.mark.xfail(
    Backend.get() == BackendName.PYSEQ,
    reason="PUT using PYSEQ does not provide immutable data",
)
def test_data_immutability():
    data = [1, 2, 3]
    object_ref = unidist.put(data)

    data[0] = 111
    assert_equal(object_ref, [1, 2, 3])



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

from libc.stdint cimport uint8_t, uintptr_t, int64_t

cdef extern from "memory.cpp" nogil:
    pass

# Declare the class with cdef
cdef extern from "memory.h" namespace "unidist" nogil:
    void parallel_memcopy(uint8_t* dst, 
                            const uint8_t* src, 
                            int64_t nbytes, 
                            uintptr_t block_size, 
                            int num_threads)

    void fill(int64_t *buff, int64_t size, int64_t value)




# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Object storage related functionality."""

from unidist.core.backends.common.data_id import DataID


class ObjectStore:
    """Class that stores objects and provides access to these."""

    __instance = None

    def __init__(self):
        if ObjectStore.__instance is None:
            self.store = dict()

    def __repr__(self):
        return f"Object store: {self.store}"

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``ObjectStore``.

        Returns
        -------
        unidist.core.backends.pyseq.core.object_store.ObjectStore
        """
        if cls.__instance is None:
            cls.__instance = ObjectStore()
        return cls.__instance

    def put(self, data, data_id=None):
        """
        Put `data` to internal dictionary.

        Parameters
        ----------
        data : object
            Data to be put.
        data_id : unidist.core.backends.common.data_id.DataID, optional
            An ID of data. If it isn't provided, will be created automatically.

        Returns
        -------
        unidist.core.backends.common.data_id.DataID
            An ID of object in internal dictionary.
        """
        data_id = DataID() if data_id is None else data_id

        self.store[data_id] = data
        return data_id

    def get(self, data_ids):
        """
        Get object(s) associated with `data_ids` from the internal dictionary.

        Parameters
        ----------
        data_ids : unidist.core.backends.common.data_id.DataID or list
            ID(s) of object(s) to get data from.

        Returns
        -------
        object
            A Python object.
        """
        is_list = isinstance(data_ids, list)
        if not is_list:
            data_ids = [data_ids]
        if not all(isinstance(data_id, DataID) for data_id in data_ids):
            raise ValueError(
                "`data_ids` must either be a data ID or a list of data IDs."
            )

        def check_exception(value):
            if isinstance(value, Exception):
                raise value
            return value

        values = [check_exception(self.store[data_id]) for data_id in data_ids]

        return values if is_list else values[0]



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Config entities specific for MPI backend which can be used for unidist behavior tuning."""

from unidist.config.parameter import EnvironmentVariable, ExactStr


class MpiSpawn(EnvironmentVariable, type=bool):
    """Whether to enable MPI spawn or not."""

    default = True
    varname = "UNIDIST_MPI_SPAWN"


class MpiHosts(EnvironmentVariable, type=ExactStr):
    """
    MPI hosts to run unidist on.

    Notes
    -----
    This variable is only used if a program is run in Controller/Worker model.
    """

    varname = "UNIDIST_MPI_HOSTS"


class MpiPickleThreshold(EnvironmentVariable, type=int):
    """
    Minimum buffer size for serialization with pickle 5 protocol.

    Notes
    -----
    If the shared object store is enabled, ``MpiSharedObjectStoreThreshold`` takes
    precedence on this configuration value and the threshold gets overridden.
    It is done intentionally to prevent multiple copies when putting an object
    into the local object store or into the shared object store.
    Data copy happens once when doing in-band serialization in depend on the threshold.
    In some cases output of a remote task can take up the memory of the task arguments.
    If those arguments are placed in the shared object store, this location should not be overwritten
    while output is being used, otherwise the output value may be corrupted.
    """

    default = 1024**2 // 4  # 0.25 MiB
    varname = "UNIDIST_MPI_PICKLE_THRESHOLD"

    @classmethod
    def get(cls) -> int:
        """
        Get minimum buffer size for serialization with pickle 5 protocol.

        Returns
        -------
        int
        """
        if MpiSharedObjectStore.get():
            mpi_pickle_threshold = MpiSharedObjectStoreThreshold.get()
            cls.put_value_source(MpiSharedObjectStoreThreshold.get_value_source())
        else:
            mpi_pickle_threshold = super().get()
        return mpi_pickle_threshold


class MpiBackoff(EnvironmentVariable, type=float):
    """
    Backoff time for preventing the "busy wait" in loops exchanging messages.

    Notes
    -----
    Use it carefully and set to a value different from the default
    in depend on a specific workload because this may slightly improve
    performance or, in contrary, deteriorate it.
    """

    default = 0.0001
    varname = "UNIDIST_MPI_BACKOFF"


class MpiLog(EnvironmentVariable, type=bool):
    """Whether to enable logging for MPI backend or not."""

    default = False
    varname = "UNIDIST_MPI_LOG"


class MpiSharedObjectStore(EnvironmentVariable, type=bool):
    """Whether to enable shared object store or not."""

    default = False
    varname = "UNIDIST_MPI_SHARED_OBJECT_STORE"


class MpiSharedObjectStoreMemory(EnvironmentVariable, type=int):
    """How many bytes of memory to start the shared object store with."""

    varname = "UNIDIST_MPI_SHARED_OBJECT_STORE_MEMORY"


class MpiSharedServiceMemory(EnvironmentVariable, type=int):
    """How many bytes of memory to start the shared service memory with."""

    varname = "UNIDIST_MPI_SHARED_SERVICE_MEMORY"


class MpiSharedObjectStoreThreshold(EnvironmentVariable, type=int):
    """Minimum size of data to put into the shared object store."""

    default = 10**5  # 100 KB
    varname = "UNIDIST_MPI_SHARED_OBJECT_STORE_THRESHOLD"


class MpiRuntimeEnv:
    """
    Runtime environment for MPI worker processes.

    Notes
    -----
    This config doesn't have a respective environment variable as
    it is much more convenient to set a config value using the config API
    but not through the environment variable.
    """

    # Possible options for a runtime environment to set
    env_vars = "env_vars"
    # Config value
    _value = {}

    @classmethod
    def put(cls, value):
        """
        Set config value.

        Parameters
        ----------
        value : dict
            Config value to set.
        """
        if any([True for option in value if option != cls.env_vars]):
            raise NotImplementedError(
                "Any option other than environment variables is not supported yet."
            )
        cls._value = value

    @classmethod
    def get(cls):
        """
        Get config value.

        Returns
        -------
        dict
        """
        return cls._value



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""An implementation of ``RemoteFunction`` interface using Python Multiprocessing backend."""

import unidist.core.backends.pymp.core as mp
from unidist.core.base.object_ref import ObjectRef
from unidist.core.base.remote_function import RemoteFunction


class PyMpRemoteFunction(RemoteFunction):
    """
    The class that implements the interface in ``RemoteFunction`` using Python Multiprocessing backend.

    Parameters
    ----------
    function : callable
        A function to be called remotely.
    num_cpus : int
        The number of CPUs to reserve for the remote function.
    num_returns : int
        The number of ``ObjectRef``-s returned by the remote function invocation.
    resources : dict
        Custom resources to reserve for the remote function.
    """

    def __init__(self, function, num_cpus, num_returns, resources):
        self._remote_function = function
        self._num_cpus = num_cpus
        self._num_returns = 1 if num_returns is None else num_returns
        self._resources = resources

    def _remote(self, *args, num_cpus=None, num_returns=None, resources=None, **kwargs):
        """
        Execute `self._remote_function` in a worker process.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed in the `self._remote_function`.
        num_cpus : int, optional
            The number of CPUs to reserve for the remote function.
        num_returns : int, optional
            The number of ``ObjectRef``-s returned by the remote function invocation.
        resources : dict, optional
            Custom resources to reserve for the remote function.
        **kwargs : dict
            Keyword arguments to be passed in the `self._remote_function`.

        Returns
        -------
        ObjectRef, list or None
            Type of returns depends on `num_returns` value:

            * if `num_returns == 1`, ``ObjectRef`` will be returned.
            * if `num_returns > 1`, list of ``ObjectRef``-s will be returned.
            * if `num_returns == 0`, ``None`` will be returned.
        """
        if num_cpus is not None or self._num_cpus is not None:
            raise NotImplementedError(
                "'num_cpus' is not supported yet by Python Multiprocessing backend."
            )
        if resources is not None or self._resources is not None:
            raise NotImplementedError(
                "'resources' is not supported yet by Python Multiprocessing backend."
            )

        if num_returns is None:
            num_returns = self._num_returns

        data_ids = mp.submit(
            self._remote_function, *args, num_returns=num_returns, **kwargs
        )

        if num_returns == 1:
            return ObjectRef(data_ids)
        elif num_returns > 1:
            return [ObjectRef(data_id) for data_id in data_ids]
        elif num_returns == 0:
            return None



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

import unidist.core.backends.mpi.core.common as common

logger = common.get_logger("async_operations", "async_operations.log")


class AsyncOperations:
    """
    Class that stores MPI async communication handlers.

    Class holds a reference to sending data to prolong data lifetime during send operation.
    """

    __instance = None

    def __init__(self):
        # I-prefixed mpi call handlers
        self._send_async_handlers = []

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``AsyncOperations``.

        Returns
        -------
        AsyncOperations
        """
        if cls.__instance is None:
            cls.__instance = AsyncOperations()
        return cls.__instance

    def extend(self, handlers_list):
        """
        Extend internal list with `handler_list`.

        Parameters
        ----------
        handler_list : list
            A list of pairs with handler and data reference.
        """
        self._send_async_handlers.extend(handlers_list)

    def check(self):
        """Check all MPI async send requests readiness and remove a reference to sending data."""

        def is_ready(handler):
            is_ready = handler.Test()
            if is_ready:
                logger.debug("CHECK ASYNC HANDLER {} - ready".format(handler))
            else:
                logger.debug("CHECK ASYNC HANDLER {} - not ready".format(handler))
            return is_ready

        # tup[0] - mpi async send handler object
        self._send_async_handlers[:] = [
            tup for tup in self._send_async_handlers if not is_ready(tup[0])
        ]

    def finish(self):
        """Finish all MPI async send requests."""
        for handler, _ in self._send_async_handlers:
            logger.debug("WAIT ASYNC HANDLER {}".format(handler))
            handler.Wait()
        self._send_async_handlers.clear()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Utilities used to initialize Dask execution backend."""

from unidist.config import (
    CpuCount,
    DaskMemoryLimit,
    IsDaskCluster,
    DaskSchedulerAddress,
)


def initialize_dask():
    """
    Initialize the Dask execution backend.

    Notes
    -----
    Number of workers for Dask Client is equal to number of CPUs used by the backend.
    """
    from distributed.client import get_client

    try:
        get_client()
    except ValueError:
        from distributed.client import Client

        num_cpus = CpuCount.get()
        memory_limit = DaskMemoryLimit.get()
        is_cluster = IsDaskCluster.get()
        scheduler_address = DaskSchedulerAddress.get()
        worker_memory_limit = memory_limit if memory_limit else "auto"

        if is_cluster:
            Client(address=scheduler_address)
        else:
            Client(
                n_workers=num_cpus,
                threads_per_worker=1,
                memory_limit=worker_memory_limit,
            )



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""High-level API."""

from .api import (
    init,
    is_initialized,
    shutdown,
    remote,
    put,
    get,
    wait,
    is_object_ref,
    get_ip,
    num_cpus,
    cluster_resources,
)
from . import _version

__all__ = [
    "init",
    "is_initialized",
    "shutdown",
    "remote",
    "put",
    "get",
    "wait",
    "is_object_ref",
    "get_ip",
    "num_cpus",
    "cluster_resources",
]

__version__ = _version.get_versions()["version"]



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""An implementation of ``Backend`` interface using Dask."""

from collections import defaultdict

from distributed.client import get_client, wait
from distributed.utils import get_ip

from unidist.core.backends.dask.actor import DaskActor
from unidist.core.backends.dask.remote_function import DaskRemoteFunction
from unidist.core.base.backend import Backend


class DaskBackend(Backend):
    """The class that implements the interface in ``Backend`` using Dask."""

    @staticmethod
    def make_remote_function(function, num_cpus, num_returns, resources):
        """
        Define a remote function.

        function : callable
            Function to be a remote function.
        num_cpus : int
            The number of CPUs to reserve for the remote function.
        num_returns : int
            The number of ``ObjectRef``-s returned by the remote function invocation.
        resources : dict
            Custom resources to reserve for the remote function.

        Returns
        -------
        DaskRemoteFunction
        """
        return DaskRemoteFunction(function, num_cpus, num_returns, resources)

    @staticmethod
    def make_actor(cls, num_cpus, resources):
        """
        Define an actor class.

        cls : object
            Class to be an actor class.
        num_cpus : int
            The number of CPUs to reserve for the lifetime of the actor.
        resources : dict
            Custom resources to reserve for the lifetime of the actor.

        Returns
        -------
        DaskActor
            The actor class type to create.
        list
            The list of arguments for ``DaskActor`` constructor.
        """
        return DaskActor, [cls, num_cpus, resources]

    @staticmethod
    def get(futures):
        """
        Get a remote object or a list of remote objects
        from distributed memory.

        Parameters
        ----------
        futures : distributed.client.Future or a list of distributed.client.Future-s
            Dask Future or a list of Dask Future objects to get data from.

        Returns
        -------
        object
            A Python object or a list of Python objects.
        """
        return (
            get_client().gather(futures)
            if isinstance(futures, list)
            else futures.result()
        )

    @staticmethod
    def put(data):
        """
        Put `data` into distributed memory.

        Parameters
        ----------
        data : object
            Data to be put.

        Returns
        -------
        distributed.client.Future
            Dask Future matching to data.
        """
        return get_client().scatter(data, hash=False)

    @staticmethod
    def wait(futures, num_returns=1):
        """
        Wait until `futures` are finished.

        This method returns two lists. The first list consists of
        futures that correspond to objects that completed computations.
        The second list corresponds to the rest of the futures (which may or may not be ready).

        Parameters
        ----------
        futures : distributed.client.Future or list
            Dask Future or list of Dask Futures to be waited.
        num_returns : int, default: 1
            The number of Dask Futures that should be returned as ready.

        Returns
        -------
        tuple
            Two lists of futures that are ready and list of the remaining futures.
        """
        done, not_done = wait(futures, return_when="FIRST_COMPLETED")
        while len(done) < num_returns:
            extra_done, not_done = wait(not_done, return_when="FIRST_COMPLETED")
            done.update(extra_done)

        done, not_done = list(done), list(not_done)
        while len(done) != num_returns:
            last = done.pop()
            not_done.append(last)

        return done, not_done

    @staticmethod
    def get_ip():
        """
        Get node IP address.

        Returns
        -------
        str
            Node IP address.
        """
        return get_ip()

    @staticmethod
    def num_cpus():
        """
        Get the number of CPUs used by the execution backend.

        Returns
        -------
        int
        """
        return len(get_client().ncores())

    @staticmethod
    def cluster_resources():
        """
        Get resources of Dask cluster.

        Returns
        -------
        dict
            Dictionary with cluster nodes info in the form
            `{"node_ip0": {"CPU": x0}, "node_ip1": {"CPU": x1}, ...}`.
        """
        client = get_client()
        cluster_resources = defaultdict(lambda: {"CPU": 0})
        for worker_info in client.scheduler_info()["workers"].values():
            cluster_resources[worker_info["host"]]["CPU"] += worker_info["nthreads"]

        localhost = "127.0.0.1"
        if localhost in cluster_resources:
            cluster_resources[get_ip()] = cluster_resources.pop(localhost)
        return dict(cluster_resources)

    @staticmethod
    def shutdown():
        """Shutdown Dask execution backend."""
        client = get_client()
        client.shutdown()

    @staticmethod
    def is_initialized():
        """
        Check if Dask backend has already been initialized.

        Returns
        -------
        bool
            True or False.
        """
        try:
            get_client()
        except ValueError:
            return False
        else:
            return True



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""An implementation of ``Backend`` interface using MPI."""

import socket

from unidist.config import CpuCount
import unidist.core.backends.mpi.core as mpi
from unidist.core.backends.mpi.actor import MPIActor
from unidist.core.backends.mpi.remote_function import MPIRemoteFunction
from unidist.core.base.backend import Backend

# Static variable for `get_ip` function
host = None


class MPIBackend(Backend):
    """The class that implements the interface in ``Backend`` using MPI."""

    @staticmethod
    def make_remote_function(function, num_cpus, num_returns, resources):
        """
        Define a remote function.

        function : callable
            Function to be a remote function.
        num_cpus : int
            The number of CPUs to reserve for the remote function.
        num_returns : int
            The number of ``ObjectRef``-s returned by the remote function invocation.
        resources : dict
            Custom resources to reserve for the remote function.

        Returns
        -------
        MPIRemoteFunction
        """
        return MPIRemoteFunction(function, num_cpus, num_returns, resources)

    @staticmethod
    def make_actor(cls, num_cpus, resources):
        """
        Define an actor class.

        cls : object
            Class to be an actor class.
        num_cpus : int
            The number of CPUs to reserve for the lifetime of the actor.
        resources : dict
            Custom resources to reserve for the lifetime of the actor.

        Returns
        -------
        MPIActor
            The actor class type to create.
        list
            The list of arguments for ``MPIActor`` constructor.
        """
        return MPIActor, [cls, num_cpus, resources]

    @staticmethod
    def get(data_ids):
        """
        Get a remote object or a list of remote objects
        from distributed memory.

        Parameters
        ----------
        data_ids : unidist.core.backends.common.data_id.DataID or list
            ``DataID`` or a list of ``DataID`` objects to get data from.

        Returns
        -------
        object
            A Python object or a list of Python objects.
        """
        return mpi.get(data_ids)

    @staticmethod
    def put(data):
        """
        Put `data` into distributed memory.

        Parameters
        ----------
        data : object
            Data to be put.

        Returns
        -------
        unidist.core.backends.common.data_id.DataID
            ``DataID`` matching to data.
        """
        return mpi.put(data)

    @staticmethod
    def wait(data_ids, num_returns=1):
        """
        Wait until `data_ids` are finished.

        This method returns two lists. The first list consists of
        data IDs that correspond to objects that completed computations.
        The second list corresponds to the rest of the data IDs (which may or may not be ready).

        Parameters
        ----------
        data_ids : unidist.core.backends.common.data_id.DataID or list
            ``DataID`` or list of ``DataID``-s to be waited.
        num_returns : int, default: 1
            The number of ``DataID``-s that should be returned as ready.

        Returns
        -------
        tuple
            List of data IDs that are ready and list of the remaining data IDs.
        """
        return mpi.wait(data_ids, num_returns=num_returns)

    @staticmethod
    def get_ip():
        """
        Get node IP address.

        Returns
        -------
        str
            Node IP address.
        """
        global host
        if host is None:
            hostname = socket.gethostname()
            host = socket.gethostbyname(hostname)
        return host

    @staticmethod
    def num_cpus():
        """
        Get the number of CPUs used by the execution backend.

        Returns
        -------
        int
        """
        return CpuCount.get()

    @staticmethod
    def cluster_resources():
        """
        Get resources of MPI cluster.

        Returns
        -------
        dict
            Dictionary with cluster nodes info in the form
            `{"node_ip0": {"CPU": x0}, "node_ip1": {"CPU": x1}, ...}`.
        """
        return mpi.cluster_resources()

    @staticmethod
    def shutdown():
        """Shutdown MPI execution backend."""
        mpi.shutdown()

    @staticmethod
    def is_initialized():
        """
        Check if MPI backend has already been initialized.

        Returns
        -------
        bool
            True or False.
        """
        return mpi.is_initialized()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""`GarbageCollector` functionality."""

import time

import unidist.core.backends.mpi.core.common as common
import unidist.core.backends.mpi.core.communication as communication
from unidist.core.backends.mpi.core.async_operations import AsyncOperations
from unidist.core.backends.mpi.core.serialization import SimpleDataSerializer
from unidist.core.backends.mpi.core.local_object_store import LocalObjectStore


logger = common.get_logger("utils", "utils.log")


class GarbageCollector:
    """
    Class that tracks deleted data IDs and cleans worker's object storage.

    Parameters
    ----------
    local_store : unidist.core.backends.mpi.core.local_object_store
        Reference to the local object storage.

    Notes
    -----
    Cleanup relies on internal threshold settings.
    """

    def __init__(self, local_store):
        # Cleanup frequency settings
        self._cleanup_counter = 1
        self._cleanup_threshold = 5
        self._time_threshold = 1  # seconds
        self._timestamp = 0  # seconds
        # Cleanup list of tuple(owner_rank, data_number)
        self._cleanup_list = []
        self._cleanup_list_threshold = 10
        # Reference to the global object store
        self._local_store = local_store
        # Task submitted counter
        self._task_counter = 0

    def _send_cleanup_request(self, cleanup_list):
        """
        Send a list of data IDs to be deleted for each worker to cleanup local storages.

        Parameters
        ----------
        cleanup_list : list
            List of data IDs.
        """
        logger.debug(f"Send cleanup list - {cleanup_list}")
        mpi_state = communication.MPIState.get_instance()
        # Cache serialized list of data IDs
        s_cleanup_list = SimpleDataSerializer().serialize_pickle(cleanup_list)
        async_operations = AsyncOperations.get_instance()
        for rank_id in mpi_state.workers + mpi_state.monitor_processes:
            if rank_id != mpi_state.global_rank:
                h_list = communication.isend_serialized_operation(
                    mpi_state.global_comm,
                    common.Operation.CLEANUP,
                    s_cleanup_list,
                    rank_id,
                )
                async_operations.extend(h_list)

    def increment_task_counter(self):
        """
        Track task submission number.

        Notes
        -----
        For cleanup purpose.
        """
        self._task_counter += 1

    def collect(self, data_id):
        """
        Track ID destruction.

        This ID is out of scope, it's associated data should be cleared.

        Parameters
        ----------
        data_id_metadata : tuple
            Tuple of the owner rank and data number describing a ``MpiDataID``.
        """
        self._cleanup_list.append(data_id)

    def regular_cleanup(self):
        """
        Cleanup all garbage collected IDs from local and all workers object storages.

        Cleanup triggers based on internal threshold settings.
        """
        logger.debug("Cleanup list len - {}".format(len(self._cleanup_list)))
        logger.debug(
            "Cleanup counter {}, threshold reached - {}".format(
                self._cleanup_counter,
                (self._cleanup_counter % self._cleanup_threshold) == 0,
            )
        )
        async_operations = AsyncOperations.get_instance()
        # Check completion status of previous async MPI routines
        async_operations.check()
        if len(self._cleanup_list) > self._cleanup_list_threshold:
            if self._cleanup_counter % self._cleanup_threshold == 0:
                timestamp_snapshot = time.perf_counter()
                if (timestamp_snapshot - self._timestamp) > self._time_threshold:
                    logger.debug("Cleanup counter {}".format(self._cleanup_counter))

                    mpi_state = communication.MPIState.get_instance()
                    root_monitor = mpi_state.get_monitor_by_worker_rank(
                        communication.MPIRank.ROOT
                    )
                    # Compare submitted and executed tasks
                    # We use a blocking send here because we have to wait for
                    # completion of the communication, which is necessary for the pipeline to continue.
                    communication.mpi_send_operation(
                        mpi_state.global_comm,
                        common.Operation.GET_TASK_COUNT,
                        root_monitor,
                    )
                    executed_task_counter = communication.mpi_recv_object(
                        mpi_state.global_comm,
                        root_monitor,
                    )

                    logger.debug(
                        "Submitted task count {} vs executed task count {}".format(
                            self._task_counter, executed_task_counter
                        )
                    )
                    if executed_task_counter == self._task_counter:
                        self._send_cleanup_request(self._cleanup_list)
                        self._cleanup_list.clear()
                        self._cleanup_counter += 1
                        self._timestamp = time.perf_counter()
            else:
                self._cleanup_counter += 1


garbage_collector = GarbageCollector(LocalObjectStore.get_instance())



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Utilities used to initialize Ray execution backend."""

import os
from packaging import version
import psutil
import ray
import sys
from typing import Optional
import warnings

from unidist.config import (
    CpuCount,
    RayGpuCount,
    IsRayCluster,
    RayRedisAddress,
    RayRedisPassword,
    RayObjectStoreMemory,
    ValueSource,
)

_OBJECT_STORE_TO_SYSTEM_MEMORY_RATIO = 0.6
# This constant should be in sync with the limit in ray, which is private,
# not exposed to users, and not documented:
# https://github.com/ray-project/ray/blob/4692e8d8023e789120d3f22b41ffb136b50f70ea/python/ray/_private/ray_constants.py#L57-L62
_MAC_OBJECT_STORE_LIMIT_BYTES = 2 * 2**30


def initialize_ray():
    """
    Initialize the Ray execution backend.

    Notes
    -----
    Number of workers for Ray is equal to number of CPUs used by the backend.
    """
    if not ray.is_initialized() or IsRayCluster.get():
        cluster = IsRayCluster.get()
        redis_address = RayRedisAddress.get()
        redis_password = (
            (
                ray.ray_constants.REDIS_DEFAULT_PASSWORD
                if cluster
                else RayRedisPassword.get()
            )
            if RayRedisPassword.get_value_source() == ValueSource.DEFAULT
            else RayRedisPassword.get()
        )

        if cluster:
            # We only start ray in a cluster setting for the head node.
            ray.init(
                address=redis_address or "auto",
                include_dashboard=False,
                ignore_reinit_error=True,
                _redis_password=redis_password,
            )
        else:
            object_store_memory = _get_object_store_memory()
            ray_init_kwargs = {
                "num_cpus": CpuCount.get(),
                "num_gpus": RayGpuCount.get(),
                "include_dashboard": False,
                "ignore_reinit_error": True,
                "object_store_memory": object_store_memory,
                "_redis_password": redis_password,
                "_memory": object_store_memory,
            }
            ray.init(**ray_init_kwargs)


def _get_object_store_memory() -> Optional[int]:
    """
    Get the object store memory we should start Ray with, in bytes.

    - If the ``Memory`` config variable is set, return that.
    - On Linux, take system memory from /dev/shm. On other systems use total
      virtual memory.
    - On Mac, never return more than Ray-specified upper limit.

    Returns
    -------
    Optional[int]
        The object store memory size in bytes, or None if we should use the Ray
        default.
    """
    object_store_memory = RayObjectStoreMemory.get()
    if object_store_memory is not None:
        return object_store_memory
    virtual_memory = psutil.virtual_memory().total
    if sys.platform.startswith("linux"):
        shm_fd = os.open("/dev/shm", os.O_RDONLY)
        try:
            shm_stats = os.fstatvfs(shm_fd)
            system_memory = shm_stats.f_bsize * shm_stats.f_bavail
            if system_memory / (virtual_memory / 2) < 0.99:
                warnings.warn(
                    f"The size of /dev/shm is too small ({system_memory} bytes). The required size "
                    + f"at least half of RAM ({virtual_memory // 2} bytes). Please, delete files in /dev/shm or "
                    + "increase size of /dev/shm with --shm-size in Docker. Also, you can can override the memory "
                    + "size for each Ray worker (in bytes) to the MODIN_MEMORY environment variable."
                )
        finally:
            os.close(shm_fd)
    else:
        system_memory = virtual_memory
    bytes_per_gb = 1e9
    object_store_memory = int(
        _OBJECT_STORE_TO_SYSTEM_MEMORY_RATIO
        * system_memory
        // bytes_per_gb
        * bytes_per_gb
    )
    if object_store_memory == 0:
        return None
    # Newer versions of ray don't allow us to initialize ray with object store
    # size larger than that _MAC_OBJECT_STORE_LIMIT_BYTES. It seems that
    # object store > the limit is too slow even on ray 1.0.0. However, limiting
    # the object store to _MAC_OBJECT_STORE_LIMIT_BYTES only seems to start
    # helping at ray version 1.3.0. So if ray version is at least 1.3.0, cap
    # the object store at _MAC_OBJECT_STORE_LIMIT_BYTES.
    # For background on the ray bug see:
    # - https://github.com/ray-project/ray/issues/20388
    # - https://github.com/modin-project/modin/issues/4872
    if sys.platform == "darwin" and version.parse(ray.__version__) >= version.parse(
        "1.3.0"
    ):
        object_store_memory = min(object_store_memory, _MAC_OBJECT_STORE_LIMIT_BYTES)
    return object_store_memory



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""High-level API of MPI backend."""

import os
import sys
import atexit
import signal
import asyncio
from collections import defaultdict

try:
    import mpi4py
except ImportError:
    raise ImportError(
        "Missing dependency 'mpi4py'. Use pip or conda to install it."
    ) from None

from unidist.core.backends.mpi.core.serialization import serialize_complex_data
from unidist.core.backends.mpi.core.object_store import ObjectStore
from unidist.core.backends.mpi.core.shared_object_store import SharedObjectStore
from unidist.core.backends.mpi.core.local_object_store import LocalObjectStore
from unidist.core.backends.mpi.core.controller.garbage_collector import (
    garbage_collector,
)
from unidist.core.backends.mpi.core.controller.common import (
    request_worker_data,
    push_data,
    RoundRobin,
)
import unidist.core.backends.mpi.core.common as common
import unidist.core.backends.mpi.core.communication as communication
from unidist.core.backends.mpi.core.async_operations import AsyncOperations
from unidist.config import (
    CpuCount,
    MpiSpawn,
    MpiHosts,
    ValueSource,
    MpiPickleThreshold,
    MpiBackoff,
    MpiLog,
    MpiSharedObjectStore,
    MpiSharedObjectStoreMemory,
    MpiSharedServiceMemory,
    MpiSharedObjectStoreThreshold,
    MpiRuntimeEnv,
)


# TODO: Find a way to move this after all imports
mpi4py.rc(recv_mprobe=False, initialize=False)
from mpi4py import MPI  # noqa: E402


logger = common.get_logger("api", "api.log")

# The global variable is responsible for if MPI backend has already been initialized
is_mpi_initialized = False
# The global variable is responsible for if MPI backend has already been shutdown
is_mpi_shutdown = False

# This should be in line with https://docs.python.org/3/library/sys.html#sys.flags
_PY_FLAGS_MAP = {
    "debug": "d",
    "inspect": "i",
    "interactive": "i",
    "isolated": "I",
    "optimize": "O",
    "dont_write_bytecode": "B",
    "no_user_site": "s",
    "no_site": "S",
    "ignore_environment": "E",
    "verbose": "v",
    "bytes_warning": "b",
    "quiet": "q",
    "hash_randomization": "R",
    "safe_path": "P",
    # The flags below are handled separately using sys._xoptions.
    # See more in https://docs.python.org/3/library/sys.html#sys._xoptions
    # 'dev_mode': 'Xdev',
    # 'utf8_mode': 'Xutf8',
    # 'int_max_str_digits': 'Xint_max_str_digits',
}


def _get_py_flags():
    """
    Get a list of the flags passed in to python.

    Returns
    -------
    list

    Notes
    -----
    This function is used to get the python flags
    in order to pass them to the workers initialization.
    """
    args = []
    for flag, opt in _PY_FLAGS_MAP.items():
        val = getattr(sys.flags, flag, 0)
        # We do not want workers to get into interactive mode
        # so the value should be 0
        val = val if opt[0] != "i" else 0
        if val > 0:
            args.append("-" + opt * val)
    for opt in sys.warnoptions:
        args.append("-W" + opt)
    sys_xoptions = getattr(sys, "_xoptions", {})
    for opt, val in sys_xoptions.items():
        args.append("-X" + opt if val is True else "-X" + opt + "=" + val)
    return args


def init():
    """
    Initialize MPI processes.

    Notes
    -----
    When initialization collect the MPI cluster topology.
    """
    is_init = MPI.Is_initialized()
    if is_init:
        thread_level = MPI.Query_thread()
        if thread_level != MPI.THREAD_MULTIPLE:
            raise RuntimeError(
                f"MPI backend supports {MPI.THREAD_MULTIPLE} thread level only, got {thread_level}"
            )
    else:
        thread_level = MPI.Init_thread()
        if thread_level < MPI.THREAD_MULTIPLE:
            raise RuntimeError(
                f"MPI backend supports {MPI.THREAD_MULTIPLE} thread level only, "
                f"but installed MPI version uses {thread_level}."
                "Please use a thread-safe MPI implementation"
            )

    runtime_env = MpiRuntimeEnv.get()

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    parent_comm = MPI.Comm.Get_parent()

    # Path to dynamically spawn MPI processes.
    # If a requirement is not met, processes have been started with mpiexec -n <N>, where N > 1.
    if rank == 0 and parent_comm == MPI.COMM_NULL and MpiSpawn.get():
        args = _get_py_flags()
        args += ["-c"]
        py_str = [
            "import unidist",
            "import unidist.config as cfg",
            "cfg.Backend.put('mpi')",
        ]
        if MpiSpawn.get_value_source() != ValueSource.DEFAULT:
            py_str += [f"cfg.MpiSpawn.put({MpiSpawn.get()})"]
        if MpiHosts.get_value_source() != ValueSource.DEFAULT:
            py_str += [f"cfg.MpiHosts.put('{MpiHosts.get()}')"]
        if CpuCount.get_value_source() != ValueSource.DEFAULT:
            py_str += [f"cfg.CpuCount.put({CpuCount.get()})"]
        if MpiPickleThreshold.get_value_source() != ValueSource.DEFAULT:
            py_str += [f"cfg.MpiPickleThreshold.put({MpiPickleThreshold.get()})"]
        if MpiBackoff.get_value_source() != ValueSource.DEFAULT:
            py_str += [f"cfg.MpiBackoff.put({MpiBackoff.get()})"]
        if MpiLog.get_value_source() != ValueSource.DEFAULT:
            py_str += [f"cfg.MpiLog.put({MpiLog.get()})"]
        if MpiSharedObjectStore.get_value_source() != ValueSource.DEFAULT:
            py_str += [f"cfg.MpiSharedObjectStore.put({MpiSharedObjectStore.get()})"]
        if MpiSharedObjectStoreMemory.get_value_source() != ValueSource.DEFAULT:
            py_str += [
                f"cfg.MpiSharedObjectStoreMemory.put({MpiSharedObjectStoreMemory.get()})"
            ]
        if MpiSharedServiceMemory.get_value_source() != ValueSource.DEFAULT:
            py_str += [
                f"cfg.MpiSharedServiceMemory.put({MpiSharedServiceMemory.get()})"
            ]
        if MpiSharedObjectStoreThreshold.get_value_source() != ValueSource.DEFAULT:
            py_str += [
                f"cfg.MpiSharedObjectStoreThreshold.put({MpiSharedObjectStoreThreshold.get()})"
            ]
        if runtime_env:
            py_str += [f"cfg.MpiRuntimeEnv.put({runtime_env})"]
            env_vars = ["import os"]
            for option, value in runtime_env.items():
                if option == MpiRuntimeEnv.env_vars:
                    for env_var_name, env_var_value in value.items():
                        env_vars += [
                            f"os.environ['{env_var_name}'] = '{env_var_value}'"
                        ]
            py_str += env_vars
        py_str += ["unidist.init()"]
        py_str = "; ".join(py_str)
        args += [py_str]

        cpu_count = CpuCount.get()
        hosts = MpiHosts.get()
        info = MPI.Info.Create()
        lib_version = MPI.Get_library_version()
        if "Intel" in lib_version:
            # To make dynamic spawn of MPI processes work properly
            # we should set this environment variable.
            # See more about Intel MPI environment variables in
            # https://www.intel.com/content/www/us/en/docs/mpi-library/developer-reference-linux/2021-8/other-environment-variables.html.
            os.environ["I_MPI_SPAWN"] = "1"

        host_list = hosts.split(",") if hosts is not None else ["localhost"]
        host_count = len(host_list)

        if common.is_shared_memory_supported():
            # +host_count to add monitor process on each host
            nprocs_to_spawn = cpu_count + host_count
        else:
            # +1 for just a single process monitor
            nprocs_to_spawn = cpu_count + 1
        if host_count > 1:
            if "Open MPI" in lib_version:
                # +1 to take into account the current root process
                # to correctly allocate slots
                slot_count = nprocs_to_spawn + 1
                slots_per_host = [
                    int(slot_count / host_count)
                    + (1 if i < slot_count % host_count else 0)
                    for i in range(host_count)
                ]
                hosts = ",".join(
                    [f"{host_list[i]}:{slots_per_host[i]}" for i in range(host_count)]
                )
                info.Set("add-host", hosts)
            else:
                info.Set("hosts", hosts)

        intercomm = MPI.COMM_SELF.Spawn(
            sys.executable,
            args,
            maxprocs=nprocs_to_spawn,
            info=info,
            root=rank,
        )
        comm = intercomm.Merge(high=False)
    else:
        if runtime_env and rank != 0:
            for option, value in runtime_env.items():
                if option == MpiRuntimeEnv.env_vars:
                    for env_var_name, env_var_value in value.items():
                        os.environ[env_var_name] = str(env_var_value)

    # path for spawned MPI processes to be merged with the parent communicator
    if parent_comm != MPI.COMM_NULL:
        comm = parent_comm.Merge(high=True)

    mpi_state = communication.MPIState.get_instance(comm)

    global is_mpi_initialized
    if not is_mpi_initialized:
        is_mpi_initialized = True

    # Initalize shared memory
    SharedObjectStore.get_instance()

    if mpi_state.is_root_process():
        atexit.register(_termination_handler)
        signal.signal(signal.SIGTERM, _termination_handler)
        signal.signal(signal.SIGINT, _termination_handler)
        # Exit the init function in root only after monitor and worker loops have started
        mpi_state.global_comm.Barrier()
        return
    elif mpi_state.is_monitor_process():
        from unidist.core.backends.mpi.core.monitor.loop import monitor_loop

        monitor_loop()
        # If the user executes a program in SPMD mode,
        # we do not want workers to continue the flow after `unidist.init()`
        # so just killing them.
        if not MpiSpawn.get():
            sys.exit()
        return
    else:
        from unidist.core.backends.mpi.core.worker.loop import worker_loop

        asyncio.run(worker_loop())
        # If the user executes a program in SPMD mode,
        # we do not want workers to continue the flow after `unidist.init()`
        # so just killing them.
        if not MpiSpawn.get():
            sys.exit()
        return


def is_initialized():
    """
    Check if MPI backend has already been initialized.

    Returns
    -------
    bool
        True or False.
    """
    global is_mpi_initialized
    return is_mpi_initialized


# TODO: cleanup before shutdown?
def shutdown():
    """
    Shutdown all MPI processes.

    Notes
    -----
    Sends cancelation operation to all workers and monitor processes.
    """
    global is_mpi_shutdown
    if not is_mpi_shutdown:
        async_operations = AsyncOperations.get_instance()
        async_operations.finish()
        mpi_state = communication.MPIState.get_instance()
        # Send shutdown commands to all ranks
        for rank_id in mpi_state.workers:
            # We use a blocking send here because we have to wait for
            # completion of the communication, which is necessary for the pipeline to continue.
            communication.mpi_send_operation(
                mpi_state.global_comm,
                common.Operation.CANCEL,
                rank_id,
            )
            logger.debug("Shutdown rank {}".format(rank_id))
        # Make sure that monitor has sent the shutdown signal to all workers.
        op_type = communication.mpi_recv_object(
            mpi_state.global_comm,
            communication.MPIRank.MONITOR,
        )
        if op_type != common.Operation.SHUTDOWN:
            raise ValueError(f"Got wrong operation type {op_type}.")
        SharedObjectStore.get_instance().finalize()
        if not MPI.Is_finalized():
            MPI.Finalize()

        logger.debug("Shutdown root")
        is_mpi_shutdown = True


def cluster_resources():
    """
    Get resources of MPI cluster.

    Returns
    -------
    dict
        Dictionary with cluster nodes info in the form
        `{"node_ip0": {"CPU": x0}, "node_ip1": {"CPU": x1}, ...}`.
    """
    mpi_state = communication.MPIState.get_instance()
    if mpi_state is None:
        raise RuntimeError("'unidist.init()' has not been called yet")

    cluster_resources = defaultdict(dict)
    for host in mpi_state.topology:
        cluster_resources[host]["CPU"] = len(
            [
                r
                for r in mpi_state.topology[host].values()
                if not mpi_state.is_root_process(r)
                and not mpi_state.is_monitor_process(r)
            ]
        )

    return dict(cluster_resources)


def put(data):
    """
    Put the data into object storage.

    Parameters
    ----------
    data : object
        Data to be put.

    Returns
    -------
    unidist.core.backends.mpi.core.common.MpiDataID
        An ID of an object in object storage.
    """
    local_store = LocalObjectStore.get_instance()
    shared_store = SharedObjectStore.get_instance()

    data_id = local_store.generate_data_id(garbage_collector)
    serialized_data = serialize_complex_data(data)
    if shared_store.is_allocated():
        shared_store.put(data_id, serialized_data)
    else:
        local_store.cache_serialized_data(data_id, serialized_data)

    logger.debug("PUT {} id".format(data_id._id))

    return data_id


def get(data_ids):
    """
    Get an object(s) associated with `data_ids` from the object storage.

    Parameters
    ----------
    data_ids : unidist.core.backends.mpi.core.common.MpiDataID or list
        An ID(s) to object(s) to get data from.

    Returns
    -------
    object
        A Python object.
    """
    object_store = ObjectStore.get_instance()
    is_list = isinstance(data_ids, list)
    if not is_list:
        data_ids = [data_ids]
    remote_data_ids = [
        data_id for data_id in data_ids if not object_store.contains(data_id)
    ]
    # Remote data gets available in the local store inside `request_worker_data`
    if remote_data_ids:
        request_worker_data(remote_data_ids)

    logger.debug("GET {} ids".format(common.unwrapped_data_ids_list(data_ids)))

    values = [object_store.get(data_id) for data_id in data_ids]

    # Initiate reference count based cleaup
    # if all the tasks were completed
    garbage_collector.regular_cleanup()

    return values if is_list else values[0]


def wait(data_ids, num_returns=1):
    """
    Wait until `data_ids` are finished.

    This method returns two lists. The first list consists of
    ``DataID``-s that correspond to objects that completed computations.
    The second list corresponds to the rest of the ``DataID``-s (which may or may not be ready).

    Parameters
    ----------
    data_ids : unidist.core.backends.mpi.core.common.MpiDataID or list
        ``DataID`` or list of ``DataID``-s to be waited.
    num_returns : int, default: 1
        The number of ``DataID``-s that should be returned as ready.

    Returns
    -------
    tuple
        List of data IDs that are ready and list of the remaining data IDs.
    """
    object_store = ObjectStore.get_instance()
    if not isinstance(data_ids, list):
        data_ids = [data_ids]
    # Since the controller should operate MpiDataID(s),
    # we use this map to retrieve and return them
    # instead of DataID(s) received from workers.
    data_id_map = dict(zip(data_ids, data_ids))
    not_ready = data_ids
    pending_returns = num_returns
    ready = []
    logger.debug("WAIT {} ids".format(common.unwrapped_data_ids_list(data_ids)))
    for data_id in not_ready.copy():
        if object_store.contains(data_id):
            ready.append(data_id)
            not_ready.remove(data_id)
            pending_returns -= 1
            if len(ready) == num_returns:
                return ready, not_ready

    operation_type = common.Operation.WAIT
    operation_data = {
        "data_ids": not_ready,
        "num_returns": pending_returns,
    }
    mpi_state = communication.MPIState.get_instance()
    root_monitor = mpi_state.get_monitor_by_worker_rank(communication.MPIRank.ROOT)
    # We use a blocking send and recv here because we have to wait for
    # completion of the communication, which is necessary for the pipeline to continue.
    communication.send_simple_operation(
        mpi_state.global_comm,
        operation_type,
        operation_data,
        root_monitor,
    )
    data = communication.mpi_recv_object(
        mpi_state.global_comm,
        root_monitor,
    )
    ready.extend(data["ready"])
    not_ready = data["not_ready"]
    # We have to retrieve and return MpiDataID(s)
    # in order for the controller to operate them in further operations.
    ready = [data_id_map[data_id] for data_id in ready]
    not_ready = [data_id_map[data_id] for data_id in not_ready]

    # Initiate reference count based cleaup
    # if all the tasks were completed
    garbage_collector.regular_cleanup()

    return ready, not_ready


def submit(task, *args, num_returns=1, **kwargs):
    """
    Execute function on a worker process.

    Parameters
    ----------
    task : callable
        Function to be executed in the worker.
    *args : iterable
        Positional arguments to be passed in the `task`.
    num_returns : int, default: 1
        Number of results to be returned from `task`.
    **kwargs : dict
        Keyword arguments to be passed in the `task`.

    Returns
    -------
    unidist.core.backends.mpi.core.common.MpiDataID or list or None
        Type of returns depends on `num_returns` value:

        * if `num_returns == 1`, ``DataID`` will be returned.
        * if `num_returns > 1`, list of ``DataID``-s will be returned.
        * if `num_returns == 0`, ``None`` will be returned.
    """
    # Initiate reference count based cleanup
    # if all the tasks were completed
    garbage_collector.regular_cleanup()

    dest_rank = RoundRobin.get_instance().schedule_rank()

    local_store = LocalObjectStore.get_instance()
    output_ids = local_store.generate_output_data_id(
        dest_rank, garbage_collector, num_returns
    )

    logger.debug("REMOTE OPERATION")
    logger.debug(
        "REMOTE args to {} rank: {}".format(
            dest_rank, common.unwrapped_data_ids_list(args)
        )
    )
    logger.debug(
        "REMOTE outputs to {} rank: {}".format(
            dest_rank, common.unwrapped_data_ids_list(output_ids)
        )
    )

    push_data(dest_rank, task)
    push_data(dest_rank, args)
    push_data(dest_rank, kwargs)

    operation_type = common.Operation.EXECUTE
    operation_data = {
        "task": task,
        # tuple cannot be serialized iteratively and it will fail if some internal data cannot be serialized using Pickle
        "args": list(args),
        "kwargs": kwargs,
        "output": output_ids,
    }
    async_operations = AsyncOperations.get_instance()
    h_list, _ = communication.isend_complex_operation(
        communication.MPIState.get_instance().global_comm,
        operation_type,
        operation_data,
        dest_rank,
        is_serialized=False,
    )
    async_operations.extend(h_list)

    # Track the task execution
    garbage_collector.increment_task_counter()

    return output_ids


# ---------------------------- #
# unidist termination handling #
# ---------------------------- #


def _termination_handler():
    shutdown()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

import unidist


def materialize(object_refs):
    return (
        unidist.get(object_refs)
        if (
            all(unidist.is_object_ref(obj_ref) for obj_ref in object_refs)
            if isinstance(object_refs, list)
            else unidist.is_object_ref(object_refs)
        )
        else object_refs
    )


def assert_equal(object_refs, expected_result):
    actual_result = materialize(object_refs)
    assert (
        actual_result == expected_result
    ), f"Actual result is <{actual_result}>, but expected <{expected_result}>."


def catch_exception(object_refs, expected_exception):
    try:
        _ = materialize(object_refs)
    except Exception as e:
        if not isinstance(e, expected_exception):
            raise Exception(f"Got {type(e)}, but {expected_exception} was expected.")
    else:
        raise Exception(
            f"Exception wasn't raised, but {expected_exception} was expected."
        )


@unidist.remote
def task(x):
    return x * x


@unidist.remote
def task_return_none():
    return None


@unidist.remote(num_returns=2)
def task_multiple_returns_default(x):
    return x, x * x


@unidist.remote
def task_multiple_returns(x):
    return x, x * x


@unidist.remote
class TestActor:
    _accumulator = 0

    def __init__(self, init=0):
        self._accumulator = init

    def get_accumulator(self):
        return self._accumulator

    def _internal(self):
        self._accumulator += 1

    def task(self, x):
        self._accumulator += x
        self._internal()
        return self._accumulator

    def task_return_none(self):
        return None

    def multiple_returns(self, x):
        return x, self._accumulator


@unidist.remote
class TestAsyncActor:
    _accumulator = 0

    def __init__(self, init=0):
        self._accumulator = init

    async def get_accumulator(self):
        return self._accumulator

    def _internal(self):
        self._accumulator += 1

    async def task(self, x):
        self._accumulator += x
        self._internal()
        return self._accumulator

    async def task_return_none(self):
        return None

    async def multiple_returns(self, x):
        return x, self._accumulator



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Config entities which can be used for common unidist behavior tuning."""

from .envvars import Backend, CpuCount

__all__ = ["Backend", "CpuCount"]



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""MPI communication interfaces."""

from collections import defaultdict
import socket
import time
import warnings

try:
    import mpi4py
except ImportError:
    raise ImportError(
        "Missing dependency 'mpi4py'. Use pip or conda to install it."
    ) from None

from unidist.config import MpiBackoff
from unidist.core.backends.mpi.core.serialization import (
    SimpleDataSerializer,
    serialize_complex_data,
    deserialize_complex_data,
)
import unidist.core.backends.mpi.core.common as common
from unidist.config.backends.mpi.envvars import MpiSpawn, MpiHosts

# TODO: Find a way to move this after all imports
mpi4py.rc(recv_mprobe=False, initialize=False)
from mpi4py import MPI  # noqa: E402
from mpi4py.util import pkl5  # noqa: E402


# Logger configuration
logger = common.get_logger("communication", "communication.log")
is_logger_header_printed = False


def log_operation(op_type, status):
    """
    Log a communication between worker processes.

    Parameters
    ----------
    op_type : unidist.core.backends.mpi.core.common.Operation
        Operation type.
    status : mpi4py.MPI.Status
        Represents the status of a reception operation.
    """
    global is_logger_header_printed
    logger_op_name_len = 15
    logger_worker_count = MPIState.get_instance().global_size

    # write header on first worker
    if (
        not is_logger_header_printed
        and MPIState.get_instance().global_rank == MPIRank.FIRST_WORKER
    ):
        worker_ids_str = "".join([f"{i}\t" for i in range(logger_worker_count)])
        logger.debug(f'#{" "*logger_op_name_len}{worker_ids_str}')
        is_logger_header_printed = True

    # Write operation to log
    source_rank = status.Get_source()
    dest_rank = MPIState.get_instance().global_rank
    op_name = common.get_op_name(op_type)
    space_after_op_name = " " * (logger_op_name_len - len(op_name))
    space_before_arrow = ".   " * (min(source_rank, dest_rank))
    space_after_arrow = "   ." * (logger_worker_count - max(source_rank, dest_rank) - 1)
    arrow_line_str = abs(dest_rank - source_rank)
    # Right arrow if dest_rank > source_rank else left
    if dest_rank > source_rank:
        arrow = f'{".---"*arrow_line_str}>'
    else:
        arrow = f'<{arrow_line_str*"---."}'
    logger.debug(
        f"{op_name}:{space_after_op_name}{space_before_arrow}{arrow}{space_after_arrow}"
    )


class MPIState:
    """
    The class holding MPI information.

    Parameters
    ----------
    comm : mpi4py.MPI.Comm
        MPI communicator.

    Attributes
    ----------
    global_comm : mpi4py.MPI.Comm
        Global MPI communicator.
    host_comm : mpi4py.MPI.Comm
        MPI subcommunicator for the current host.
    global_rank : int
        Rank of a process.
    global_size : int
        Number of processes in the global communicator.
    host : str
        IP-address of the current host.
    topology : dict
        Dictionary, containing all ranks assignments by IP-addresses in
        the form: `{"node_ip0": {"host_rank": "global_rank", ...}, ...}`.
    host_by_rank : dict
        Dictionary containing IP addresses by rank.
    monitor_processes : list
        List of ranks that are monitor processes.
    workers : list
        List of ranks that are worker processes.
    """

    __instance = None

    def __init__(self, comm):
        # attributes get actual values when MPI is initialized in `init` function
        self.global_comm = comm
        self.global_rank = comm.Get_rank()
        self.global_size = comm.Get_size()
        self.host = socket.gethostbyname(socket.gethostname())
        # `Split_type` does not work correctly in the MSMPI library in C/W model
        # so we split the communicator in a different way
        if "Microsoft MPI" in MPI.Get_library_version() and MpiSpawn.get():
            all_hosts = self.global_comm.allgather(self.host)
            self.host_comm = self.global_comm.Split(all_hosts.index(self.host))
        else:
            self.host_comm = self.global_comm.Split_type(MPI.COMM_TYPE_SHARED)

        host_rank = self.host_comm.Get_rank()
        # Get topology of MPI cluster.
        cluster_info = self.global_comm.allgather(
            (self.host, self.global_rank, host_rank)
        )
        self.topology = defaultdict(dict)
        self.host_by_rank = defaultdict(None)
        for host, global_rank, host_rank in cluster_info:
            self.topology[host][host_rank] = global_rank
            self.host_by_rank[global_rank] = host

        mpi_hosts = MpiHosts.get()
        if mpi_hosts is not None and MpiSpawn.get():
            host_list = mpi_hosts.split(",")
            host_count = len(host_list)

            # check running hosts
            if self.is_root_process() and len(self.topology.keys()) > host_count:
                warnings.warn(
                    "The number of running hosts is greater than that specified in the UNIDIST_MPI_HOSTS. "
                    + "If you want to run the program on a host other than the local one, specify the appropriate parameter for `mpiexec` "
                    + "(`--host` for OpenMPI and `--hosts` for Intel MPI or MPICH)."
                )
            if self.is_root_process() and len(self.topology.keys()) < host_count:
                warnings.warn(
                    "The number of running hosts is less than that specified in the UNIDIST_MPI_HOSTS. "
                    + "Check the `mpiexec` option to distribute processes between hosts."
                )
        if common.is_shared_memory_supported():
            self.monitor_processes = []
            for host in self.topology:
                if len(self.topology[host]) >= 2:
                    self.monitor_processes.append(self.topology[host][MPIRank.MONITOR])
                elif self.is_root_process():
                    raise ValueError(
                        "When using shared object store, each host must contain at least 2 processes, "
                        "since one of them will be a service monitor."
                    )
        else:
            self.monitor_processes = [MPIRank.MONITOR]

        self.workers = []
        for host in self.topology:
            self.workers.extend(
                [
                    rank
                    for rank in self.topology[host].values()
                    if not self.is_root_process(rank)
                    and not self.is_monitor_process(rank)
                ]
            )

    @classmethod
    def get_instance(cls, *args):
        """
        Get instance of this class.

        Parameters
        ----------
        *args : tuple
            Positional arguments to create the instance.
            See the constructor's docstring on the arguments.

        Returns
        -------
        MPIState
        """
        if cls.__instance is None and args:
            cls.__instance = MPIState(*args)
        return cls.__instance

    def is_root_process(self, rank=None):
        """
        Check if the rank is root process.

        Parameters
        ----------
        rank : int, optional
            The rank to be checked.
            If ``None``, the current rank is to be checked.

        Returns
        -------
        bool
            True or False.
        """
        if rank is None:
            rank = self.global_rank
        return rank == MPIRank.ROOT

    def is_monitor_process(self, rank=None):
        """
        Check if the rank is a monitor process.

        Parameters
        ----------
        rank : int, optional
            The rank to be checked.
            If ``None``, the current rank is to be checked.

        Returns
        -------
        bool
            True or False.
        """
        if rank is None:
            rank = self.global_rank
        return rank in self.monitor_processes

    def get_monitor_by_worker_rank(self, rank=None):
        """
        Get the monitor process rank for the host that includes this rank.

        Parameters
        ----------
        rank : int, optional
            The global rank to search for a monitor process.

        Returns
        -------
        int
            Rank of a monitor process.
        """
        if not common.is_shared_memory_supported():
            return MPIRank.MONITOR

        if rank is None:
            rank = self.global_rank
        host = self.host_by_rank[rank]
        if host is None:
            raise ValueError("Unknown rank of workers")

        return self.topology[host][MPIRank.MONITOR]


class MPIRank:
    """Class that describes rank assignment."""

    ROOT = 0
    MONITOR = 1
    FIRST_WORKER = 2


# ---------------------------- #
# Main communication utilities #
# ---------------------------- #


def mpi_send_operation(comm, op_type, dest_rank):
    """
    Send an operation type to another MPI rank in a blocking way.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    op_type : unidist.core.backends.mpi.core.common.Operation
        An operation type to send.
    dest_rank : int
        Target MPI process to transfer data.

    Notes
    -----
    * This blocking send is used when we have to wait for completion of the communication,
      which is necessary for the pipeline to continue, or when the receiver is waiting for a result.
      Otherwise, use non-blocking ``mpi_isend_operation``.
    * The special tag is used for this communication, namely, ``common.MPITag.OPERATION``.
    """
    comm.send(op_type, dest=dest_rank, tag=common.MPITag.OPERATION)


def mpi_send_object(comm, data, dest_rank, tag=common.MPITag.OBJECT):
    """
    Send a Python object to another MPI rank in a blocking way.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    data : object
        Data to send.
    dest_rank : int
        Target MPI process to transfer data.
    tag : common.MPITag, default: common.MPITag.OBJECT
        Message tag.

    Notes
    -----
    * This blocking send is used when we have to wait for completion of the communication,
      which is necessary for the pipeline to continue, or when the receiver is waiting for a result.
      Otherwise, use non-blocking ``mpi_isend_object``.
    * The special tag is used for this communication, namely, ``common.MPITag.OBJECT``.
    """
    comm.send(data, dest=dest_rank, tag=tag)


def mpi_isend_operation(comm, op_type, dest_rank):
    """
    Send an operation type to another MPI rank in a non-blocking way.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    op_type : unidist.core.backends.mpi.core.common.Operation
        An operation type to send.
    dest_rank : int
        Target MPI process to transfer data.

    Returns
    -------
    object
        A handler to MPI_Isend communication result.

    Notes
    -----
    The special tag is used for this communication, namely, ``common.MPITag.OPERATION``.
    """
    return comm.isend(op_type, dest=dest_rank, tag=common.MPITag.OPERATION)


def mpi_isend_object(comm, data, dest_rank):
    """
    Send a Python object to another MPI rank in a non-blocking way.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    data : object
        Data to send.
    dest_rank : int
        Target MPI process to transfer data.

    Returns
    -------
    object
        A handler to MPI_Isend communication result.

    Notes
    -----
    The special tag is used for this communication, namely, ``common.MPITag.OBJECT``.
    """
    return comm.isend(data, dest=dest_rank, tag=common.MPITag.OBJECT)


def mpi_recv_operation(comm):
    """
    Worker receive operation type interface.

    Busy waits to avoid contention. Receives data from any source.

    Parameters
    ----------
    comm : object
        MPI communicator object.

    Returns
    -------
    unidist.core.backends.mpi.core.common.Operation
        Operation type.
    int
        Source rank.

    Notes
    -----
    The special tag is used for this communication, namely, ``common.MPITag.OPERATION``.
    """
    backoff = MpiBackoff.get()
    status = MPI.Status()
    source = MPI.ANY_SOURCE
    tag = common.MPITag.OPERATION
    while not comm.iprobe(source=source, tag=tag, status=status):
        time.sleep(backoff)
    source = status.source
    tag = status.tag
    op_type = comm.recv(buf=None, source=source, tag=tag, status=status)
    log_operation(op_type, status)
    return op_type, status.Get_source()


def mpi_iprobe_recv_object(comm, tag=common.MPITag.OBJECT):
    """
    Receive an object of a standard Python data type from any source.

    The source rank gets available from `iprobe`.

    Parameters
    ----------
    comm : mpi4py.MPI.Comm
        MPI communicator.
    tag : common.MPITag, default: common.MPITag.OBJECT
        Message tag.

    Returns
    -------
    object
        Received data from the source rank.
    int
        Source rank.
    """
    backoff = MpiBackoff.get()
    status = MPI.Status()
    source = MPI.ANY_SOURCE
    while not comm.iprobe(source=source, tag=tag, status=status):
        time.sleep(backoff)
    source = status.source
    data = comm.recv(source=source, tag=tag, status=status)
    return data, source


def mpi_recv_object(comm, source_rank):
    """
    Receive an object of a standard Python data type.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    source_rank : int
        Source MPI process to receive data from.

    Returns
    -------
    object
        Received data object from another MPI process.

    Notes
    -----
    * De-serialization is a simple pickle.load in this case.
    * The special tag is used for this communication, namely, ``common.MPITag.OBJECT``.
    """
    return comm.recv(source=source_rank, tag=common.MPITag.OBJECT)


def mpi_send_buffer(comm, buffer, dest_rank, data_type=MPI.CHAR, buffer_size=None):
    """
    Send buffer object to another MPI rank in a blocking way.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    buffer : object
        Buffer object to send.
    dest_rank : int
        Target MPI process to transfer buffer.
    data_type : MPI.Datatype, default: MPI.CHAR
        MPI data type for sending data.
    buffer_size: int, default: None
        Buffer size in bytes. Send an additional message with a buffer size to prepare another process to receive if `buffer_size` is not None.

    Notes
    -----
    * This blocking send is used when we have to wait for completion of the communication,
      which is necessary for the pipeline to continue, or when the receiver is waiting for a result.
      Otherwise, use non-blocking ``mpi_isend_buffer``.
    * The special tags are used for this communication, namely,
      ``common.MPITag.OBJECT`` and ``common.MPITag.BUFFER``.
    """
    if buffer_size:
        comm.send(buffer_size, dest=dest_rank, tag=common.MPITag.OBJECT)
    else:
        buffer_size = len(buffer)
    # Maximum block size MPI is able to send/recv
    block_size = pkl5._bigmpi.blocksize
    partitions = list(range(0, buffer_size, block_size))
    partitions.append(buffer_size)
    num_partitions = len(partitions)
    with pkl5._bigmpi as bigmpi:
        for i in range(num_partitions):
            if i + 1 < num_partitions:
                comm.Send(
                    bigmpi(buffer[partitions[i] : partitions[i + 1]]),
                    dest=dest_rank,
                    tag=common.MPITag.BUFFER,
                )


def mpi_isend_buffer(comm, buffer_size, buffer, dest_rank):
    """
    Send buffer object to another MPI rank in a non-blocking way.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    buffer_size : int
        Buffer size in bytes.
    buffer : object
        Buffer object to send.
    dest_rank : int
        Target MPI process to transfer data.

    Returns
    -------
    object
        A handler to MPI_Isend communication result.

    Notes
    -----
    The special tags are used for this communication, namely,
    ``common.MPITag.OBJECT`` and ``common.MPITag.BUFFER``.
    """
    requests = []
    h1 = comm.isend(buffer_size, dest=dest_rank, tag=common.MPITag.OBJECT)
    requests.append((h1, None))
    # Maximum block size MPI is able to send/recv
    block_size = pkl5._bigmpi.blocksize
    partitions = list(range(0, buffer_size, block_size))
    partitions.append(buffer_size)
    num_partitions = len(partitions)
    with pkl5._bigmpi as bigmpi:
        for i in range(num_partitions):
            if i + 1 < num_partitions:
                h2 = comm.Isend(
                    bigmpi(buffer[partitions[i] : partitions[i + 1]]),
                    dest=dest_rank,
                    tag=common.MPITag.BUFFER,
                )
                requests.append((h2, buffer))

    return requests


def mpi_recv_buffer(comm, source_rank, result_buffer=None):
    """
    Receive data buffer.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    source_rank : int
        Communication event source rank.
    result_buffer : object, default: None
        The array to be filled. If `result_buffer` is None, the buffer size will be requested and the necessary buffer created.

    Returns
    -------
    object
        Array buffer or serialized object.

    Notes
    -----
    The special tags are used for this communication, namely,
    ``common.MPITag.OBJECT`` and ``common.MPITag.BUFFER``.
    """
    if result_buffer is None:
        buf_size = comm.recv(source=source_rank, tag=common.MPITag.OBJECT)
        result_buffer = bytearray(buf_size)
    else:
        buf_size = len(result_buffer)
    # Maximum block size MPI is able to send/recv
    block_size = pkl5._bigmpi.blocksize
    partitions = list(range(0, buf_size, block_size))
    partitions.append(buf_size)
    num_partitions = len(partitions)
    with pkl5._bigmpi as bigmpi:
        for i in range(num_partitions):
            if i + 1 < num_partitions:
                tmp_buffer = bytearray(partitions[i + 1] - partitions[i])
                comm.Recv(
                    bigmpi(tmp_buffer), source=source_rank, tag=common.MPITag.BUFFER
                )
                result_buffer[partitions[i] : partitions[i + 1]] = tmp_buffer

    return result_buffer


def mpi_busy_wait_recv(comm, source_rank):
    """
    Wait for receive operation result in a custom busy wait loop.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    source_rank : int
        Source MPI process to receive data.

    Returns
    -------
    object
        Received data.

    Notes
    -----
    The special tag is used for this communication, namely, ``common.MPITag.OBJECT``.
    """
    backoff = MpiBackoff.get()
    req_handle = comm.irecv(source=source_rank, tag=common.MPITag.OBJECT)
    while True:
        status, data = req_handle.test()
        if status:
            return data
        else:
            time.sleep(backoff)


# --------------------------------- #
# Communication operation functions #
# --------------------------------- #


def _send_complex_data_impl(comm, s_data, raw_buffers, dest_rank, info_package):
    """
    Send already serialized complex data.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    s_data : bytearray
        Serialized data as bytearray.
    raw_buffers : list
        Pickle buffers list, out-of-band data collected with pickle 5 protocol.
    dest_rank : int
        Target MPI process to transfer data.
    info_package : unidist.core.backends.mpi.core.common.MetadataPackage
        Required information to deserialize data on a receiver side.

    Notes
    -----
    The special tags are used for this communication, namely,
    ``common.MPITag.OBJECT`` and ``common.MPITag.BUFFER``.
    """
    # wrap to dict for sending and correct deserialization of the object by the recipient
    comm.send(dict(info_package), dest=dest_rank, tag=common.MPITag.OBJECT_BLOCKING)
    with pkl5._bigmpi as bigmpi:
        comm.Send(bigmpi(s_data), dest=dest_rank, tag=common.MPITag.BUFFER)
        for sbuf in raw_buffers:
            comm.Send(bigmpi(sbuf), dest=dest_rank, tag=common.MPITag.BUFFER)


def send_complex_data(comm, data, dest_rank, is_serialized=False):
    """
    Send the data that consists of different user provided complex types, lambdas and buffers in a blocking way.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    data : object
        Data object to send.
    dest_rank : int
        Target MPI process to transfer data.
    is_serialized : bool, default: False
        `data` is already serialized or not.

    Returns
    -------
    dict
        Serialized data for caching purpose.

    Notes
    -----
    This blocking send is used when we have to wait for completion of the communication,
    which is necessary for the pipeline to continue, or when the receiver is waiting for a result.
    Otherwise, use non-blocking ``isend_complex_data``.
    """
    if is_serialized:
        s_data = data["s_data"]
        raw_buffers = data["raw_buffers"]
        buffer_count = data["buffer_count"]
        # pop `data_id` out of the dict because it will be send as part of metadata package
        data_id = data.pop("id")
        serialized_data = data
    else:
        data_id = data["id"]
        serialized_data = serialize_complex_data(data)
        s_data = serialized_data["s_data"]
        raw_buffers = serialized_data["raw_buffers"]
        buffer_count = serialized_data["buffer_count"]

    info_package = common.MetadataPackage.get_local_info(
        data_id, len(s_data), [len(sbuf) for sbuf in raw_buffers], buffer_count
    )
    # MPI communication
    _send_complex_data_impl(comm, s_data, raw_buffers, dest_rank, info_package)

    # For caching purpose
    return serialized_data


def _isend_complex_data_impl(comm, s_data, raw_buffers, dest_rank, info_package):
    """
    Send serialized complex data.

    Non-blocking asynchronous interface.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    s_data : object
        A serialized msgpack data.
    raw_buffers : list
        A list of pickle buffers.
    dest_rank : int
        Target MPI process to transfer data.
    info_package : unidist.core.backends.mpi.core.common.MetadataPackage
        Required information to deserialize data on a receiver side.

    Returns
    -------
    list
        A list of pairs, ``MPI_Isend`` handler and associated data to send.

    Notes
    -----
    * The special tags are used for this communication, namely,
      ``common.MPITag.OBJECT`` and ``common.MPITag.BUFFER``.
    """
    handlers = []
    # wrap to dict for sending and correct deserialization of the object by the recipient
    h1 = comm.isend(dict(info_package), dest=dest_rank, tag=common.MPITag.OBJECT)
    handlers.append((h1, None))
    with pkl5._bigmpi as bigmpi:
        h2 = comm.Isend(bigmpi(s_data), dest=dest_rank, tag=common.MPITag.BUFFER)
        handlers.append((h2, s_data))
        for sbuf in raw_buffers:
            h_sbuf = comm.Isend(bigmpi(sbuf), dest=dest_rank, tag=common.MPITag.BUFFER)
            handlers.append((h_sbuf, sbuf))

    return handlers


def isend_complex_data(comm, data, dest_rank, is_serialized=False):
    """
    Send the data that consists of different user provided complex types, lambdas and buffers in a non-blocking way.

    Non-blocking asynchronous interface.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    data : object
        Data object to send.
    dest_rank : int
        Target MPI process to transfer data.
    is_serialized : bool, default: False
        `operation_data` is already serialized or not.

    Returns
    -------
    list
        A list of pairs, ``MPI_Isend`` handler and associated data to send.
    object
        A serialized msgpack data.
    list
        A list of pickle buffers.
    list
        A list of buffers amount for each object.

    Notes
    -----
    The special tags are used for this communication, namely,
    ``common.MPITag.OBJECT`` and ``common.MPITag.BUFFER``.
    """
    if is_serialized:
        s_data = data["s_data"]
        raw_buffers = data["raw_buffers"]
        buffer_count = data["buffer_count"]
        # pop `data_id` out of the dict because it will be send as part of metadata package
        data_id = data.pop("id")
        info_package = common.MetadataPackage.get_local_info(
            data_id, len(s_data), [len(sbuf) for sbuf in raw_buffers], buffer_count
        )
    else:
        serialized_data = serialize_complex_data(data)
        s_data = serialized_data["s_data"]
        raw_buffers = serialized_data["raw_buffers"]
        buffer_count = serialized_data["buffer_count"]
        info_package = common.MetadataPackage.get_task_info(
            len(s_data), [len(sbuf) for sbuf in raw_buffers], buffer_count
        )

    # MPI communication
    handlers = _isend_complex_data_impl(
        comm, s_data, raw_buffers, dest_rank, info_package
    )

    return handlers, s_data, raw_buffers, buffer_count


def recv_complex_data(comm, source_rank, info_package):
    """
    Receive the data that may consist of different user provided complex types, lambdas and buffers.

    The data is de-serialized from received buffer.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    source_rank : int
        Source MPI process to receive data from.
    info_package : unidist.core.backends.mpi.core.common.MetadataPackage
        Required information to deserialize data.

    Returns
    -------
    object
        Received data object from another MPI process.

    Notes
    -----
    The special tags are used for this communication, namely,
    ``common.MPITag.OBJECT`` and ``common.MPITag.BUFFER``.
    """
    msgpack_buffer = bytearray(info_package["s_data_len"])
    buffer_count = info_package["buffer_count"]
    raw_buffers = list(map(bytearray, info_package["raw_buffers_len"]))
    with pkl5._bigmpi as bigmpi:
        comm.Recv(bigmpi(msgpack_buffer), source=source_rank, tag=common.MPITag.BUFFER)
        for rbuf in raw_buffers:
            comm.Recv(bigmpi(rbuf), source=source_rank, tag=common.MPITag.BUFFER)

    return deserialize_complex_data(msgpack_buffer, raw_buffers, buffer_count)


# ---------- #
# Public API #
# ---------- #


def send_simple_operation(comm, operation_type, operation_data, dest_rank):
    """
    Send an operation type and standard Python data types in a blocking way.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    operation_type : unidist.core.backends.mpi.core.common.Operation
        Operation message type.
    operation_data : object
        Data object to send.
    dest_rank : int
        Target MPI process to transfer data.

    Notes
    -----
    * This blocking send is used when we have to wait for completion of the communication,
      which is necessary for the pipeline to continue, or when the receiver is waiting for a result.
      Otherwise, use non-blocking ``isend_simple_operation``.
    * Serialization of the data to be sent takes place just using ``pickle.dump`` in this case.
    * The special tags are used for this communication, namely,
      ``common.MPITag.OPERATION`` and ``common.MPITag.OBJECT``.
    """
    # Send operation type
    mpi_send_operation(comm, operation_type, dest_rank)
    # Send the details of a communication request
    mpi_send_object(comm, operation_data, dest_rank)


def isend_simple_operation(comm, operation_type, operation_data, dest_rank):
    """
    Send an operation type and standard Python data types in a non-blocking way.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    operation_type : unidist.core.backends.mpi.core.common.Operation
        Operation message type.
    operation_data : object
        Data object to send.
    dest_rank : int
        Target MPI process to transfer data.

    Returns
    -------
    list
        A list of pairs, ``MPI_Isend`` handler and associated data to send.

    Notes
    -----
    * Serialization of the data to be sent takes place just using ``pickle.dump`` in this case.
    * The special tags are used for this communication, namely,
      ``common.MPITag.OPERATION`` and ``common.MPITag.OBJECT``.
    """
    # Send operation type
    handlers = []
    h1 = mpi_isend_operation(comm, operation_type, dest_rank)
    handlers.append((h1, operation_type))
    # Send the details of a communication request
    h2 = mpi_isend_object(comm, operation_data, dest_rank)
    handlers.append((h2, operation_data))
    return handlers


def isend_complex_operation(
    comm, operation_type, operation_data, dest_rank, is_serialized=False
):
    """
    Send operation and data that consists of different user provided complex types, lambdas and buffers.

    Non-blocking asynchronous interface.
    The data is serialized with ``unidist.core.backends.mpi.core.ComplexDataSerializer``.
    Function works with already serialized data.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    operation_type : ``unidist.core.backends.mpi.core.common.Operation``
        Operation message type.
    operation_data : object
        Data object to send.
    dest_rank : int
        Target MPI process to transfer data.
    is_serialized : bool
        `operation_data` is already serialized or not.
        - `operation_data` is always serialized for data
        that has already been saved into the object store.
        - `operation_data` is always not serialized
        for sending a task or an actor (actor method).

    Returns
    -------
    list and dict
        Async handlers list and serialization data dict for caching purpose.

    Notes
    -----
    The special tags are used for this communication, namely,
    ``common.MPITag.OPERATION``, ``common.MPITag.OBJECT`` and ``common.MPITag.BUFFER``.
    """
    # Send operation type
    handlers = []
    h1 = mpi_isend_operation(comm, operation_type, dest_rank)
    handlers.append((h1, None))

    # Send operation data
    if is_serialized:
        # Send already serialized data
        s_data = operation_data["s_data"]
        raw_buffers = operation_data["raw_buffers"]
        buffer_count = operation_data["buffer_count"]
        # pop `data_id` out of the dict because it will be send as part of metadata package
        data_id = operation_data.pop("id")
        info_package = common.MetadataPackage.get_local_info(
            data_id, len(s_data), [len(sbuf) for sbuf in raw_buffers], buffer_count
        )
        h2_list = _isend_complex_data_impl(
            comm, s_data, raw_buffers, dest_rank, info_package
        )
        handlers.extend(h2_list)
    else:
        # Serialize and send the data
        h2_list, s_data, raw_buffers, buffer_count = isend_complex_data(
            comm, operation_data, dest_rank
        )
        handlers.extend(h2_list)
    return handlers, {
        "s_data": s_data,
        "raw_buffers": raw_buffers,
        "buffer_count": buffer_count,
    }


def isend_serialized_operation(comm, operation_type, operation_data, dest_rank):
    """
    Send operation and serialized simple data.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    operation_type : unidist.core.backends.mpi.core.common.Operation
        Operation message type.
    operation_data : object
        Data object to send.
    dest_rank : int
        Target MPI process to transfer data.

    Returns
    -------
    list
        A list of pairs, ``MPI_Isend`` handler and associated data to send.

    Notes
    -----
    The special tags are used for this communication, namely,
    ``common.MPITag.OPERATION``, ``common.MPITag.OBJECT`` and ``common.MPITag.BUFFER``.
    """
    handlers = []
    # Send operation type
    h1 = mpi_isend_operation(comm, operation_type, dest_rank)
    handlers.append((h1, operation_type))
    # Send the details of a communication request
    h2_list = mpi_isend_buffer(comm, len(operation_data), operation_data, dest_rank)
    handlers.extend(h2_list)
    return handlers


def recv_serialized_data(comm, source_rank):
    """
    Receive serialized data buffer.

    The data is de-serialized with ``unidist.core.backends.mpi.core.SimpleDataSerializer``.

    Parameters
    ----------
    comm : object
        MPI communicator object.
    source_rank : int
        Source MPI process to receive data from.

    Returns
    -------
    object
        Received de-serialized data object from another MPI process.

    Notes
    -----
    The special tags are used for this communication, namely,
    ``common.MPITag.OBJECT`` and ``common.MPITag.BUFFER``.
    """
    s_buffer = mpi_recv_buffer(comm, source_rank)
    return SimpleDataSerializer().deserialize_pickle(s_buffer)


def send_reserve_operation(comm, data_id, data_size):
    """
    Reserve shared memory for `data_id`.

    Parameters
    ----------
    data_id : unidist.core.backends.mpi.core.common.MpiDataID
        An ID to data.
    data_size : int
        Length of a required range in shared memory.

    Returns
    -------
    dict
        Reservation info about the allocated range in shared memory.
    """
    mpi_state = MPIState.get_instance()
    operation_type = common.Operation.RESERVE_SHARED_MEMORY

    operation_data = {
        "id": data_id,
        "size": data_size,
    }
    # We use a blocking send here because we have to wait for
    # completion of the communication, which is necessary for the pipeline to continue.
    send_simple_operation(
        comm,
        operation_type,
        operation_data,
        mpi_state.get_monitor_by_worker_rank(),
    )
    return mpi_recv_object(comm, mpi_state.get_monitor_by_worker_rank())



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import functools
import inspect
import time

from unidist.core.backends.common.data_id import is_data_id
import unidist.core.backends.mpi.core.common as common
import unidist.core.backends.mpi.core.communication as communication
from unidist.core.backends.mpi.core.async_operations import AsyncOperations
from unidist.core.backends.mpi.core.object_store import ObjectStore
from unidist.core.backends.mpi.core.local_object_store import LocalObjectStore
from unidist.core.backends.mpi.core.shared_object_store import SharedObjectStore
from unidist.core.backends.mpi.core.serialization import serialize_complex_data
from unidist.core.backends.mpi.core.worker.request_store import RequestStore

mpi_state = communication.MPIState.get_instance()
# Logger configuration
# When building documentation we do not have MPI initialized so
# we use the condition to set "worker_0.log" in order to build it succesfully.
logger_name = "worker_{}".format(mpi_state.global_rank if mpi_state is not None else 0)
log_file = "{}.log".format(logger_name)
w_logger = common.get_logger(logger_name, log_file)


class TaskStore:
    """Class that stores tasks/actor-tasks that couldn't be executed due to data dependencies."""

    __instance = None

    def __init__(self):
        # Incomplete tasks list - arguments not ready yet
        self._pending_tasks_list = []
        # Incomplete actor tasks list - arguments not ready yet
        self._pending_actor_tasks_list = []
        # Event loop for executing coroutines
        self.event_loop = asyncio.get_event_loop()
        # Started async tasks
        self.background_tasks = set()

    @classmethod
    def get_instance(cls):
        """
        Get instance of ``TaskStore``.

        Returns
        -------
        TaskStore
        """
        if cls.__instance is None:
            cls.__instance = TaskStore()
        return cls.__instance

    def put(self, request):
        """
        Save task execution request for later processing.

        Some data dependencies are not resolved yet.

        Parameters
        ----------
        request : dict
            Task execution request with arguments.
        """
        self._pending_tasks_list.append(request)

    def put_actor(self, request):
        """
        Save actor task execution request for later processing.

        Some data dependencies are not resolved yet.

        Parameters
        ----------
        request : dict
            Actor task execution request with arguments.
        """
        self._pending_actor_tasks_list.append(request)

    def check_pending_tasks(self):
        """
        Check a list of pending task execution requests and process all ready tasks.

        Task is ready if all data dependencies are resolved.
        """
        w_logger.debug("Check pending tasks")

        if self._pending_tasks_list:
            updated_list = []
            for request in self._pending_tasks_list:
                pending_request = self.process_task_request(request)
                if pending_request:
                    updated_list.append(pending_request)
            self._pending_tasks_list = updated_list

    def clear_pending_tasks(self):
        """
        Clear a list of pending task execution requests.
        """
        w_logger.debug("Clear pending tasks")

        self._pending_tasks_list.clear()

    def check_pending_actor_tasks(self):
        """
        Check a list of pending actor task execution requests and process all ready tasks.

        Task is ready if all data dependencies are resolved.
        """
        w_logger.debug("Check pending actor tasks")

        if self._pending_actor_tasks_list:
            updated_list = []
            for request in self._pending_actor_tasks_list:
                pending_request = self.process_task_request(request)
                if pending_request:
                    updated_list.append(pending_request)
            self._pending_actor_tasks_list = updated_list

    def clear_pending_actor_tasks(self):
        """
        Clear a list of pending actor task execution requests.
        """
        w_logger.debug("Clear pending actor tasks")

        self._pending_actor_tasks_list.clear()

    def request_worker_data(self, dest_rank, data_id):
        """
        Send GET operation with data request to destination worker.

        Parameters
        ----------
        dest_rank : int
            Rank number to request data from.
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            `data_id` associated data to request.

        Notes
        -----
        Request is asynchronous, no wait for the data.
        """
        w_logger.debug(
            "Request {} id from {} worker rank".format(data_id._id, dest_rank)
        )

        operation_type = common.Operation.GET
        operation_data = {
            "source": communication.MPIState.get_instance().global_rank,
            "id": data_id,
            "is_blocking_op": False,
        }
        async_operations = AsyncOperations.get_instance()
        h_list = communication.isend_simple_operation(
            communication.MPIState.get_instance().global_comm,
            operation_type,
            operation_data,
            dest_rank,
        )
        async_operations.extend(h_list)

        # Save request in order to prevent massive communication during pending task checks
        RequestStore.get_instance().put(data_id, dest_rank, RequestStore.DATA)

    def unwrap_local_data_id(self, arg):
        """
        Inspect argument and get the ID associated data from the local object store if available.

        If the object store is missing this data ID, request the data from another worker.

        Parameters
        ----------
        arg : object or unidist.core.backends.mpi.core.common.MpiDataID
            Data ID or object to inspect.

        Returns
        -------
        tuple
            Same value or data ID associated data and special flag.

        Notes
        -----
        The function returns the success status of data materialization attempt as a flag.
        If the data ID could not be resolved, the function returns ``True``.
        """
        if is_data_id(arg):
            local_store = LocalObjectStore.get_instance()
            object_store = ObjectStore.get_instance()
            if object_store.contains(arg):
                value = object_store.get(arg)
                # Data is already local or was pushed from master
                return value, False
            elif local_store.contains_data_owner(arg):
                if not RequestStore.get_instance().is_data_already_requested(arg):
                    # Request the data from an owner worker
                    owner_rank = local_store.get_data_owner(arg)
                    if owner_rank != communication.MPIState.get_instance().global_rank:
                        self.request_worker_data(owner_rank, arg)
                return arg, True
            else:
                raise ValueError("DataID is missing!")
        else:
            return arg, False

    def execute_received_task(self, output_data_ids, task, args, kwargs):
        """
        Execute a task/actor-task and handle results.

        Parameters
        ----------
        output_data_ids : list of unidist.core.backends.mpi.core.common.MpiDataID
            A list of output data IDs to store the results in local object store.
        task : callable
            Function to be executed.
        args : iterable
            Positional arguments to be passed in the `task`.
        kwargs : dict
            Keyword arguments to be passed in the `task`.

        Notes
        -----
        Exceptions are stored in output data IDs as value.
        """
        local_store = LocalObjectStore.get_instance()
        shared_store = SharedObjectStore.get_instance()
        completed_data_ids = []
        # Note that if a task is coroutine,
        # the local store or the shared store will contain output data
        # only once the task is complete.
        if inspect.iscoroutinefunction(task):

            async def execute():
                try:
                    w_logger.debug("- Start task execution -")

                    for arg in args:
                        if isinstance(arg, Exception):
                            raise arg
                    for value in kwargs.values():
                        if isinstance(value, Exception):
                            raise value

                    start = time.perf_counter()

                    # Execute user task
                    f_to_execute = functools.partial(task, *args, **kwargs)
                    output_values = await f_to_execute()

                    w_logger.info(
                        "Task evaluation time: {}".format(time.perf_counter() - start)
                    )
                    w_logger.debug("- End task execution -")

                except Exception as e:
                    w_logger.debug("Exception - {}".format(e))

                    if (
                        isinstance(output_data_ids, (list, tuple))
                        and len(output_data_ids) > 1
                    ):
                        for output_id in output_data_ids:
                            local_store.put(output_id, e)
                    else:
                        local_store.put(output_data_ids, e)
                else:
                    if output_data_ids is not None:
                        if (
                            isinstance(output_data_ids, (list, tuple))
                            and len(output_data_ids) > 1
                        ):
                            completed_data_ids = [None] * len(output_data_ids)
                            for idx, (output_id, value) in enumerate(
                                zip(output_data_ids, output_values)
                            ):
                                serialized_data = serialize_complex_data(value)
                                local_store.put(output_id, value)
                                if (
                                    shared_store.is_allocated()
                                    and shared_store.should_be_shared(serialized_data)
                                ):
                                    shared_store.put(output_id, serialized_data)
                                else:
                                    local_store.cache_serialized_data(
                                        output_id, serialized_data
                                    )
                                completed_data_ids[idx] = output_id
                        else:
                            serialized_data = serialize_complex_data(output_values)
                            local_store.put(output_data_ids, output_values)
                            if (
                                shared_store.is_allocated()
                                and shared_store.should_be_shared(serialized_data)
                            ):
                                shared_store.put(output_data_ids, serialized_data)
                            else:
                                local_store.cache_serialized_data(
                                    output_data_ids, serialized_data
                                )
                            completed_data_ids = [output_data_ids]

                RequestStore.get_instance().check_pending_get_requests(output_data_ids)
                # Monitor the task execution
                # We use a blocking send here because we have to wait for
                # completion of the communication, which is necessary for the pipeline to continue.
                root_monitor = mpi_state.get_monitor_by_worker_rank(
                    communication.MPIRank.ROOT
                )
                communication.send_simple_operation(
                    communication.MPIState.get_instance().global_comm,
                    common.Operation.TASK_DONE,
                    completed_data_ids,
                    root_monitor,
                )

            async_task = asyncio.create_task(execute())
            # Add task to the set. This creates a strong reference.
            self.background_tasks.add(async_task)
            # To prevent keeping references to finished tasks forever,
            # make each task remove its own reference from the set after
            # completion.
            async_task.add_done_callback(self.background_tasks.discard)
        else:
            try:
                w_logger.debug("- Start task execution -")

                for arg in args:
                    if isinstance(arg, Exception):
                        raise arg
                for value in kwargs.values():
                    if isinstance(value, Exception):
                        raise value

                start = time.perf_counter()

                # Execute user task
                output_values = task(*args, **kwargs)

                w_logger.info(
                    "Task evaluation time: {}".format(time.perf_counter() - start)
                )
                w_logger.debug("- End task execution -")

            except Exception as e:
                w_logger.debug("Exception - {}".format(e))

                if (
                    isinstance(output_data_ids, (list, tuple))
                    and len(output_data_ids) > 1
                ):
                    for output_id in output_data_ids:
                        local_store.put(output_id, e)
                else:
                    local_store.put(output_data_ids, e)
            else:
                if output_data_ids is not None:
                    if (
                        isinstance(output_data_ids, (list, tuple))
                        and len(output_data_ids) > 1
                    ):
                        completed_data_ids = [None] * len(output_data_ids)
                        for idx, (output_id, value) in enumerate(
                            zip(output_data_ids, output_values)
                        ):
                            serialized_data = serialize_complex_data(value)
                            local_store.put(output_id, value)
                            if (
                                shared_store.is_allocated()
                                and shared_store.should_be_shared(serialized_data)
                            ):
                                shared_store.put(output_id, serialized_data)
                            else:
                                local_store.cache_serialized_data(
                                    output_id, serialized_data
                                )
                            completed_data_ids[idx] = output_id
                    else:
                        serialized_data = serialize_complex_data(output_values)
                        local_store.put(output_data_ids, output_values)
                        if (
                            shared_store.is_allocated()
                            and shared_store.should_be_shared(serialized_data)
                        ):
                            shared_store.put(output_data_ids, serialized_data)
                        else:
                            local_store.cache_serialized_data(
                                output_data_ids, serialized_data
                            )
                        completed_data_ids = [output_data_ids]
            RequestStore.get_instance().check_pending_get_requests(output_data_ids)
            # Monitor the task execution.
            # We use a blocking send here because we have to wait for
            # completion of the communication, which is necessary for the pipeline to continue.
            root_monitor = mpi_state.get_monitor_by_worker_rank(
                communication.MPIRank.ROOT
            )
            communication.send_simple_operation(
                communication.MPIState.get_instance().global_comm,
                common.Operation.TASK_DONE,
                completed_data_ids,
                root_monitor,
            )

    def process_task_request(self, request):
        """
        Parse request data and execute the task if possible.

        Data dependencies should be resolved for task execution.

        Parameters
        ----------
        request : dict
            Task related data (args, function, output).

        Returns
        -------
        dict or None
            Same request if the task couldn`t be executed, otherwise ``None``.
        """
        object_store = ObjectStore.get_instance()
        # Parse request
        task = request["task"]
        # Remote function here is a data id so we have to retrieve it from the storage,
        # whereas actor method is already materialized in the worker loop.
        if is_data_id(task):
            task = object_store.get(task)
        args = request["args"]
        kwargs = request["kwargs"]
        output_ids = request["output"]

        w_logger.debug("REMOTE task: {}".format(task))
        w_logger.debug("REMOTE args: {}".format(common.unwrapped_data_ids_list(args)))
        w_logger.debug(
            "REMOTE kwargs: {}".format(common.unwrapped_data_ids_list(kwargs))
        )
        w_logger.debug(
            "REMOTE outputs: {}".format(common.unwrapped_data_ids_list(output_ids))
        )

        # DataID -> real data
        args, is_pending = common.materialize_data_ids(args, self.unwrap_local_data_id)
        kwargs, is_kw_pending = common.materialize_data_ids(
            kwargs, self.unwrap_local_data_id
        )

        w_logger.debug("Is pending - {}".format(is_pending))

        if is_pending or is_kw_pending:
            request["args"] = args
            request["kwargs"] = kwargs
            return request
        else:
            self.execute_received_task(output_ids, task, args, kwargs)
            if output_ids is not None:
                RequestStore.get_instance().check_pending_get_requests(output_ids)
                RequestStore.get_instance().check_pending_wait_requests(output_ids)
            return None

    def __del__(self):
        self.event_loop.close()



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Config entities specific for MPI backend which can be used for unidist behavior tuning."""

from .envvars import (
    MpiSpawn,
    MpiHosts,
    MpiPickleThreshold,
    MpiBackoff,
    MpiLog,
    MpiSharedObjectStore,
    MpiSharedObjectStoreMemory,
    MpiSharedServiceMemory,
    MpiSharedObjectStoreThreshold,
    MpiRuntimeEnv,
)

__all__ = [
    "MpiSpawn",
    "MpiHosts",
    "MpiPickleThreshold",
    "MpiBackoff",
    "MpiLog",
    "MpiSharedObjectStore",
    "MpiSharedObjectStoreMemory",
    "MpiSharedServiceMemory",
    "MpiSharedObjectStoreThreshold",
    "MpiRuntimeEnv",
]



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Utilities used to initialize Python Multiprocessing execution backend."""

from unidist.config import CpuCount


def initialize_pymp():
    """
    Initialize the Python Multiprocessing execution backend.

    Notes
    -----
    Number of workers for Python Multiprocessing is equal to number of CPUs used by the backend.
    """
    from unidist.core.backends.pymp.core import init

    init(num_workers=CpuCount.get())



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Config entities specific for Ray backend which can be used for unidist behavior tuning."""

import secrets

from unidist.config.parameter import EnvironmentVariable, ExactStr


class RayGpuCount(EnvironmentVariable, type=int):
    """How many GPU devices to use during initialization of the Ray backend."""

    varname = "UNIDIST_RAY_GPUS"


class IsRayCluster(EnvironmentVariable, type=bool):
    """Whether Ray is running on pre-initialized Ray cluster."""

    varname = "UNIDIST_RAY_CLUSTER"


class RayRedisAddress(EnvironmentVariable, type=ExactStr):
    """Redis address to connect to when running in Ray cluster."""

    varname = "UNIDIST_RAY_REDIS_ADDRESS"


class RayRedisPassword(EnvironmentVariable, type=ExactStr):
    """What password to use for connecting to Redis."""

    varname = "UNIDIST_RAY_REDIS_PASSWORD"
    default = secrets.token_hex(32)


class RayObjectStoreMemory(EnvironmentVariable, type=int):
    """How many bytes of memory to start the Ray object store with."""

    varname = "UNIDIST_RAY_OBJECT_STORE_MEMORY"



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Core base object ref specific functionality."""


class ObjectRef:
    """
    A class that wraps an object ref specific for the backend.

    Parameters
    ----------
    ref : object ref
        An object ref specific for the backend.
    """

    def __init__(self, ref):
        self._ref = ref



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Actor specific functionality using MPI backend."""

import unidist.core.backends.mpi.core as mpi
from unidist.core.base.actor import Actor, ActorMethod
from unidist.core.base.object_ref import ObjectRef


class MPIActorMethod(ActorMethod):
    """
    The class implements the interface in ``ActorMethod`` using MPI backend.

    Parameters
    ----------
    cls : unidist.core.backends.mpi.core.Actor
        An actor class from which method `method_name` will be called.
    method_name : str
        The name of the method to be called.
    """

    def __init__(self, cls, method_name):
        self._cls = cls
        self._method_name = method_name
        self._num_returns = 1

    def _remote(self, *args, num_returns=None, **kwargs):
        """
        Execute `self._method_name` in a worker process.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed in the method.
        num_returns : int, optional
            Number of results to be returned. If it isn't
            provided, `self._num_returns` will be used.
        **kwargs : dict
            Keyword arguments to be passed in the method.

        Returns
        -------
        ObjectRef, list or None
            Type of returns depends on `num_returns` value:

            * if `num_returns == 1`, ``ObjectRef`` will be returned.
            * if `num_returns > 1`, list of ``ObjectRef``-s will be returned.
            * if `num_returns == 0`, ``None`` will be returned.
        """
        if num_returns is None:
            num_returns = self._num_returns

        class_method = getattr(self._cls, self._method_name)
        data_ids = class_method(*args, num_returns=num_returns, **kwargs)

        if num_returns == 1:
            return ObjectRef(data_ids)
        elif num_returns > 1:
            return [ObjectRef(ref) for ref in data_ids]
        elif num_returns == 0:
            return None


class MPIActor(Actor):
    """
    The class implements the interface in ``Actor`` using MPI backend.

    Parameters
    ----------
    cls : object
        Class to be an actor class.
    num_cpus : int
        The number of CPUs to reserve for the lifetime of the actor.
    resources : dict
        Custom resources to reserve for the lifetime of the actor.
    actor_handle : None or unidist.core.backends.mpi.core.Actor
        Used for proper serialization/deserialization via `__reduce__`.
    """

    def __init__(self, cls, num_cpus, resources, actor_handle=None):
        self._cls = cls
        self._num_cpus = num_cpus
        self._resources = resources
        self._actor_handle = actor_handle

    def _serialization_helper(self):
        """
        Helper to save the state of the object.

        This is defined to make pickling work via `__reduce__`.

        Returns
        -------
        dict
            A dictionary of the information needed to reconstruct the object.
        """
        state = {
            "cls": self._cls,
            "num_cpus": self._num_cpus,
            "resources": self._resources,
            "actor_handle": self._actor_handle,
        }
        return state

    @classmethod
    def _deserialization_helper(cls, state):
        """
        Helper to restore the state of the object.

        This is defined to make pickling work via `__reduce__`.

        Parameters
        ----------
        state : dict
            The serialized state of the object.

        Returns
        -------
        MPIActor
        """
        return cls(
            state["cls"],
            state["num_cpus"],
            state["resources"],
            actor_handle=state["actor_handle"],
        )

    def __reduce__(self):
        """
        This is defined intentionally to make pickling work correctly.

        Returns
        -------
        tuple
            Callable and arguments to be passed in to it.
        """
        state = self._serialization_helper()
        return self._deserialization_helper, (state,)

    def __getattr__(self, name):
        """
        Get the attribute `name` of the `self._cls` class.

        This methods creates the ``MPIActorMethod`` object that is responsible
        for calling method `name` of the `self._cls` class remotely.

        Parameters
        ----------
        name : str
            Name of the method to be called remotely.

        Returns
        -------
        MPIActorMethod
        """
        return MPIActorMethod(self._actor_handle, name)

    def _remote(self, *args, num_cpus=None, resources=None, **kwargs):
        """
        Create actor class, specific for MPI backend, from `self._cls`.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed in `self._cls` class constructor.
        num_cpus : int, optional
            The number of CPUs to reserve for the lifetime of the actor.
        resources : dict, optional
            Custom resources to reserve for the lifetime of the actor.
        **kwargs : dict
            Keyword arguments to be passed in `self._cls` class constructor.

        Returns
        -------
        MPIActor
        """
        if num_cpus is not None or self._num_cpus is not None:
            raise NotImplementedError("'num_cpus' is not supported yet by MPI backend.")
        if resources is not None or self._resources is not None:
            raise NotImplementedError(
                "'resources' is not supported yet by MPI backend."
            )

        self._actor_handle = mpi.Actor(self._cls, *args, **kwargs)
        return self



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Core base actor specific functionality."""

from .common import filter_arguments


class ActorMethod:
    """
    A class that is base for any actor method class specific for the backend.

    This wraps an instance of the child class, which in turn wraps
    an actor method that is meant to be remote, specific for the backend.

    Parameters
    ----------
    actor_method_cls : ActorMethod
        An instance of the ``ActorMethod`` child class.
    """

    def __init__(self, actor_method_cls):
        self._actor_method_cls = actor_method_cls

    def options(self, *args, num_returns=None, **kwargs):
        """
        Override the actor method invocation parameters.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be overrided.
        num_returns : int, optional
            The number of object refs returned by the remote function invocation.
        **kwargs : dict
            Keyword arguments to be overrided.

        Returns
        -------
        FuncWrapper
            An instance of wrapped method that a non-underscore .remote() can be called on.
        """
        actor_method_cls = self._actor_method_cls

        class FuncWrapper:
            def remote(self, *args, **kwargs):
                args, kwargs = filter_arguments(*args, **kwargs)

                return actor_method_cls._remote(
                    *args, num_returns=num_returns, **kwargs
                )

        return FuncWrapper()

    def remote(self, *args, **kwargs):
        """
        Call an actor method in a worker process.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed in the actor method.
        **kwargs : dict
            Keyword arguments to be passed in the actor method.

        Returns
        -------
        ObjectRef, list or None
        """
        args, kwargs = filter_arguments(*args, **kwargs)

        return self._actor_method_cls._remote(*args, **kwargs)


class Actor:
    """
    A class that is base for any actor class specific for the backend.

    This wraps an instance of the child class, which in turn wraps
    an actor that is meant to be remote, specific for the backend.

    Parameters
    ----------
    actor_cls : Actor
        An instance of the ``Actor`` child class.
    """

    def __init__(self, actor_cls):
        self._actor_cls = actor_cls

    def _serialization_helper(self):
        """
        Helper to save the state of the object.

        This is defined to make pickling work via `__reduce__`.

        Returns
        -------
        dict
            A dictionary of the information needed to reconstruct the object.
        """
        state = {
            "actor_cls": self._actor_cls,
        }
        return state

    @classmethod
    def _deserialization_helper(cls, state):
        """
        Helper to restore the state of the object.

        This is defined to make pickling work via `__reduce__`.

        Parameters
        ----------
        state : dict
            The serialized state of the object.

        Returns
        -------
        Actor
        """
        return cls(
            state["actor_cls"],
        )

    def __reduce__(self):
        """
        This is defined intentionally to make pickling work correctly.

        Returns
        -------
        tuple
            Callable and arguments to be passed in to it.
        """
        state = self._serialization_helper()
        return self._deserialization_helper, (state,)

    def __getattr__(self, name):
        """
        Get the attribute `name` of the actor class.

        This methods creates the ``ActorMethod`` object whose child
        is responsible for calling method `name` of the actor class remotely.

        Parameters
        ----------
        name : str
            Name of the method to be called remotely.

        Returns
        -------
        ActorMethod
        """
        actor_method_cls = getattr(self._actor_cls, name)
        return ActorMethod(actor_method_cls)


class ActorClass:
    """
    A class that serves as a actor class decorator.

    This wraps an instance of the actor class, which in turn wraps
    an actor that is meant to be remote, specific for the backend.

    Parameters
    ----------
    actor_cls : Actor
        An instance of the ``Actor`` child class.
    """

    def __init__(self, actor_cls, *actor_cls_args):
        self._actor_cls = actor_cls
        self._actor_cls_args = actor_cls_args

    def __call__(self, *args, **kwargs):
        raise TypeError(
            "Actors cannot be instantiated directly. "
            f"Instead of '{self._actor_cls_args[0].__name__}()', "
            f"use '{self._actor_cls_args[0].__name__}.remote()'."
        )

    def options(self, *args, num_cpus=None, resources=None, **kwargs):
        """
        Override the actor instantiation parameters.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be overrided.
        num_cpus : int, optional
            The number of CPUs to reserve for the lifetime of the actor.
        resources : dict, optional
            Custom resources to reserve for the lifetime of the actor.
        **kwargs : dict
            Keyword arguments to be overrided.

        Returns
        -------
        ActorWrapper
            An instance of wrapped class that a non-underscore .remote() can be called on.
        """
        actor_cls = self._actor_cls
        actor_cls_args = self._actor_cls_args

        class ActorWrapper:
            def remote(self, *args, **kwargs):
                return Actor(
                    actor_cls(*actor_cls_args)._remote(
                        *args,
                        num_cpus=num_cpus,
                        resources=resources,
                        **kwargs,
                    )
                )

        return ActorWrapper()

    def remote(self, *args, **kwargs):
        """
        Instantiate an actor class.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed in the class constructor
            that is meant to be actor.
        **kwargs : dict
            Keyword arguments to be passed in the class constructor
            that is meant to be actor.

        Returns
        -------
        Actor
        """
        actor_cls = self._actor_cls(*self._actor_cls_args)._remote(*args, **kwargs)
        return Actor(actor_cls)



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Utilities used to initialize execution backend."""

from unidist.config import Backend
from unidist.core.base.common import BackendName
from .backend import BackendProxy


def init_backend():
    """
    Initialize an execution backend.

    Notes
    -----
    The concrete execution backend can be set via
    `UNIDIST_BACKEND` environment variable or ``Backend`` config value.
    MPI backend is used by default.
    """
    backend_name = Backend.get()

    if backend_name == BackendName.MPI:
        from unidist.core.backends.mpi.backend import MPIBackend
        from unidist.core.backends.mpi.utils import initialize_mpi

        initialize_mpi()
        backend_cls = MPIBackend()
    elif backend_name == BackendName.DASK:
        import threading

        if threading.current_thread() is threading.main_thread():
            from unidist.core.backends.dask.backend import DaskBackend
            from unidist.core.backends.dask.utils import initialize_dask

            initialize_dask()
            backend_cls = DaskBackend()
    elif backend_name == BackendName.RAY:
        from unidist.core.backends.ray.backend import RayBackend
        from unidist.core.backends.ray.utils import initialize_ray

        initialize_ray()
        backend_cls = RayBackend()
    elif backend_name == BackendName.PYMP:
        from unidist.core.backends.pymp.backend import PyMpBackend
        from unidist.core.backends.pymp.utils import (
            initialize_pymp,
        )

        initialize_pymp()
        backend_cls = PyMpBackend()
    elif backend_name == BackendName.PYSEQ:
        from unidist.core.backends.pyseq.backend import PySeqBackend
        from unidist.core.backends.pyseq.utils import initialize_pyseq

        initialize_pyseq()
        backend_cls = PySeqBackend()
    else:
        raise ImportError("Unrecognized execution backend.")

    BackendProxy.get_instance(backend_cls=backend_cls)


def get_backend_proxy():
    """
    Get proxy object of the backend through which operations will be performed.

    Returns
    -------
    Backend
        The ``Backend`` instance that is considered as the proxy object.
    """
    backend = BackendProxy.get_instance()

    if backend is None:
        backend_name = Backend.get()
        if backend_name == BackendName.MPI:
            from unidist.core.backends.mpi.backend import MPIBackend

            backend_cls = MPIBackend()
        elif backend_name == BackendName.DASK:
            from unidist.core.backends.dask.backend import DaskBackend

            backend_cls = DaskBackend()
        elif backend_name == BackendName.RAY:
            from unidist.core.backends.ray.backend import RayBackend

            backend_cls = RayBackend()
        elif backend_name == BackendName.PYMP:
            from unidist.core.backends.pymp.backend import (
                PyMpBackend,
            )

            backend_cls = PyMpBackend()
        elif backend_name == BackendName.PYSEQ:
            from unidist.core.backends.pyseq.backend import PySeqBackend

            backend_cls = PySeqBackend()
        else:
            raise ValueError("Unrecognized execution backend.")

        backend = BackendProxy.get_instance(backend_cls=backend_cls)

    return backend



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""High-level API of Python Sequential backend."""


from unidist.core.backends.common.data_id import DataID
from unidist.core.backends.pyseq.core.object_store import ObjectStore


# The global variable is responsible for if Python Sequential backend has already been initialized
is_python_initialized = False


def init():
    """
    Initialize an object storage.

    Notes
    -----
    Run initialization of singleton object ``unidist.core.backends.pyseq.core.object_store.ObjectStore``.
    """
    ObjectStore.get_instance()
    global is_python_initialized
    if not is_python_initialized:
        is_python_initialized = True


def is_initialized():
    """
    Check if Python Sequential backend has already been initialized.

    Returns
    -------
    bool
        True or False.
    """
    global is_python_initialized
    return is_python_initialized


def put(data):
    """
    Put data into object storage.

    Parameters
    ----------
    data : object
        Data to be put.

    Returns
    -------
    unidist.core.backends.common.data_id.DataID
        An ID of object in object storage.
    """
    return ObjectStore.get_instance().put(data)


def get(data_ids):
    """
    Get object(s) associated with `data_ids` from the object storage.

    Parameters
    ----------
    data_ids : unidist.core.backends.common.data_id.DataID or list
        ID(s) to object(s) to get data from.

    Returns
    -------
    object
        A Python object.
    """
    return ObjectStore.get_instance().get(data_ids)


def submit(func, *args, num_returns=1, **kwargs):
    """
    Execute function.

    Parameters
    ----------
    func : callable
        Function to be executed.
    *args : iterable
        Positional arguments to be passed in the `func`.
    num_returns : int, default: 1
        Number of results to be returned from `func`.
    **kwargs : dict
        Keyword arguments to be passed in the `func`.

    Returns
    -------
    unidist.core.backends.common.data_id.DataID, list or None
        Type of returns depends on `num_returns` value:

        * if `num_returns == 1`, ``DataID`` will be returned.
        * if `num_returns > 1`, list of ``DataID``-s will be returned.
        * if `num_returns == 0`, ``None`` will be returned.
    """
    obj_store = ObjectStore.get_instance()

    materialized_args = [
        obj_store.get(arg) if isinstance(arg, DataID) else arg for arg in args
    ]
    materialized_kwargs = {
        key: obj_store.get(value) if isinstance(value, DataID) else value
        for key, value in kwargs.items()
    }

    try:
        result = func(*materialized_args, **materialized_kwargs)
    except Exception as e:
        result = [e] * num_returns if num_returns > 1 else e

    if num_returns == 0:
        data_ids = None
    elif num_returns > 1:
        data_ids = [obj_store.put(result[idx]) for idx in range(num_returns)]
    else:
        data_ids = obj_store.put(result)

    return data_ids



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""Utilities for the backends."""

from unidist import is_object_ref


def unwrap_object_refs(obj_refs):
    """Find all ``unidist.core.base.ObjectRef`` instances and unwrap underlying objects.

    Parameters
    ----------
    obj_refs : iterable or ObjectRef
        Iterable objects to transform recursively.

    Returns
    -------
    iterable or underlying object of ObjectRef
    """
    if type(obj_refs) in (list, tuple, dict):
        container = type(obj_refs)()
        for value in obj_refs:
            unwrapped_value = unwrap_object_refs(
                obj_refs[value] if isinstance(obj_refs, dict) else value
            )
            if isinstance(container, list):
                container += [unwrapped_value]
            elif isinstance(container, tuple):
                container += (unwrapped_value,)
            elif isinstance(container, dict):
                container.update({value: unwrapped_value})
        return container
    else:
        return obj_refs._ref if is_object_ref(obj_refs) else obj_refs



# Copyright (C) 2021-2023 Modin authors
#
# SPDX-License-Identifier: Apache-2.0

"""`SharedMemoryManager` functionality."""

from array import array

try:
    import mpi4py
except ImportError:
    raise ImportError(
        "Missing dependency 'mpi4py'. Use pip or conda to install it."
    ) from None

from unidist.core.backends.mpi.core import communication, common
from unidist.core.backends.mpi.core.shared_object_store import SharedObjectStore
from unidist.core.backends.mpi.utils import ImmutableDict

# TODO: Find a way to move this after all imports
mpi4py.rc(recv_mprobe=False, initialize=False)
from mpi4py import MPI  # noqa: E402


class FreeMemoryRange:
    """
    Class that helps keep track of free space in shared memory.

    Parameters
    ----------
    range_len : int
        Memory length.
    """

    def __init__(self, range_len):
        self.range = [[0, range_len]]

    def occupy(self, count=1):
        """
        Take the place of a certain length in memory.

        Parameters
        ----------
        count : int
            Required number of elements in memory.

        Returns
        -------
        int
            First index in memory.
        int
            Last index in memory.
        """
        first_index = None
        last_index = None
        for i in range(len(self.range)):
            if count <= self.range[i][1] - self.range[i][0]:
                first_index = self.range[i][0]
                last_index = first_index + count
                if self.range[i][1] == last_index:
                    self.range = self.range[:i] + self.range[i + 1 :]
                else:
                    self.range[i][0] = last_index
                break

        return first_index, last_index

    def release(self, first_index, last_index):
        """
        Free up memory space.

        Parameters
        ----------
        first_index : int
            First index in memory.
        last_index : int
            Last index in memory (not inclusive).
        """
        if len(self.range) == 0:
            self.range.append([first_index, last_index])
        elif self.range[-1][1] < first_index:
            self.range.append([first_index, last_index])
        else:
            for i in range(len(self.range)):
                if self.range[i][0] == last_index:
                    if self.range[i - 1][1] == first_index:
                        self.range[i - 1][1] = self.range[i][1]
                        self.range = self.range[:i] + self.range[i + 1 :]
                    else:
                        self.range[i][0] = first_index
                    break
                if self.range[i][1] == first_index:
                    if len(self.range) > i + 1 and self.range[i + 1][0] == last_index:
                        self.range[i + 1][0] = self.range[i][0]
                        self.range = self.range[:i] + self.range[i + 1 :]
                    else:
                        self.range[i][1] = last_index
                    break
                if self.range[i][0] > last_index:
                    self.range = (
                        self.range[:i] + [[first_index, last_index]] + self.range[i:]
                    )
                    break


class SharedMemoryManager:
    """
    Class that helps manage shared memory.
    """

    def __init__(self):
        self.shared_store = None
        if common.is_shared_memory_supported():
            self.shared_store = SharedObjectStore.get_instance()
            self._reservation_info = {}
            self.free_memory = FreeMemoryRange(self.shared_store.shared_memory_size)
            self.free_service_indexes = FreeMemoryRange(
                self.shared_store.service_info_max_count
            )
            self.pending_cleanup = []

            self.monitor_comm = None
            mpi_state = communication.MPIState.get_instance()

            monitor_group = mpi_state.global_comm.Get_group().Incl(
                mpi_state.monitor_processes
            )
            self.monitor_comm = mpi_state.global_comm.Create_group(monitor_group)

    def get(self, data_id):
        """
        Get the reservation information for the `data_id`.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.

        Returns
        -------
        dict or None
            Reservation information.

        Notes
        -----
        The `dict` is returned if a reservation has been specified, otherwise `False` is returned.
        """
        if self.shared_store is None:
            raise RuntimeError(
                "`SharedMemoryManager` cannot be used if the shared object storage is not enabled."
            )
        if data_id not in self._reservation_info:
            return None
        return self._reservation_info[data_id]

    def put(self, data_id, memory_len):
        """
        Reserve memory for the `data_id`.

        Parameters
        ----------
        data_id : unidist.core.backends.mpi.core.common.MpiDataID
            An ID to data.
        memory_len : int
            Required memory length.

        Returns
        -------
        dict
            Reservation information.
        """
        if self.shared_store is None:
            raise RuntimeError(
                "`SharedMemoryManager` cannot be used if the shared object storage is not enabled."
            )
        first_index, last_index = self.free_memory.occupy(memory_len)
        service_index, _ = self.free_service_indexes.occupy(SharedObjectStore.INFO_SIZE)
        if first_index is None:
            raise MemoryError("Overflow memory")
        if service_index is None:
            raise MemoryError("Overflow service memory")

        reservation_info = ImmutableDict(
            {
                "first_index": first_index,
                "last_index": last_index,
                "service_index": service_index,
            }
        )

        self._reservation_info[data_id] = reservation_info
        return reservation_info

    def clear(self, data_id_list):
        """
        Clear shared memory for the list of `DataID` if possible.

        Parameters
        ----------
        data_id_list : list
            List of `DataID`.
        """
        if self.shared_store is None:
            return
        cleanup_list = self.pending_cleanup + data_id_list
        self.pending_cleanup = []

        has_refs = array(
            "B",
            [
                (
                    1
                    if data_id in self._reservation_info
                    and self.shared_store.get_ref_number(
                        data_id, self._reservation_info[data_id]["service_index"]
                    )
                    > 0
                    else 0
                )
                for data_id in cleanup_list
            ],
        )

        if self.monitor_comm is not None:
            all_refs = array("B", [0] * len(has_refs))
            self.monitor_comm.Allreduce(has_refs, all_refs, MPI.MAX)
        else:
            all_refs = has_refs

        for data_id, referers in zip(cleanup_list, all_refs):
            if referers == 0:
                if data_id in self._reservation_info:
                    reservation_info = self._reservation_info[data_id]
                    self.shared_store.delete_service_info(
                        data_id, reservation_info["service_index"]
                    )
                    self.free_service_indexes.release(
                        reservation_info["service_index"],
                        reservation_info["service_index"] + SharedObjectStore.INFO_SIZE,
                    )
                    self.free_memory.release(
                        reservation_info["first_index"], reservation_info["last_index"]
                    )
                    del self._reservation_info[data_id]
            else:
                self.pending_cleanup.append(data_id)
